# Copyright 2009-2010 Gregory P. Ward
# Copyright 2009-2010 Intelerad Medical Systems Incorporated
# Copyright 2010-2011 Fog Creek Software
# Copyright 2010-2011 Unity Technologies
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

'''largefiles utility code: must not import other modules in this package.'''
from __future__ import absolute_import

import contextlib
import copy
import os
import stat

from mercurial.i18n import _
from mercurial.node import hex
from mercurial.pycompat import open

from mercurial import (
    dirstate,
    encoding,
    error,
    httpconnection,
    match as matchmod,
    pycompat,
    requirements,
    scmutil,
    sparse,
    util,
    vfs as vfsmod,
)
from mercurial.utils import hashutil

shortname = b'.hglf'
shortnameslash = shortname + b'/'
longname = b'largefiles'

# -- Private worker functions ------------------------------------------


@contextlib.contextmanager
def lfstatus(repo, value=True):
    oldvalue = getattr(repo, 'lfstatus', False)
    repo.lfstatus = value
    try:
        yield
    finally:
        repo.lfstatus = oldvalue


def getminsize(ui, assumelfiles, opt, default=10):
    lfsize = opt
    if not lfsize and assumelfiles:
        lfsize = ui.config(longname, b'minsize', default=default)
    if lfsize:
        try:
            lfsize = float(lfsize)
        except ValueError:
            raise error.Abort(
                _(b'largefiles: size must be number (not %s)\n') % lfsize
            )
    if lfsize is None:
        raise error.Abort(_(b'minimum size for largefiles must be specified'))
    return lfsize


def link(src, dest):
    """Try to create hardlink - if that fails, efficiently make a copy."""
    util.makedirs(os.path.dirname(dest))
    try:
        util.oslink(src, dest)
    except OSError:
        # if hardlinks fail, fallback on atomic copy
        with open(src, b'rb') as srcf, util.atomictempfile(dest) as dstf:
            for chunk in util.filechunkiter(srcf):
                dstf.write(chunk)
        os.chmod(dest, os.stat(src).st_mode)


def usercachepath(ui, hash):
    """Return the correct location in the "global" largefiles cache for a file
    with the given hash.
    This cache is used for sharing of largefiles across repositories - both
    to preserve download bandwidth and storage space."""
    return os.path.join(_usercachedir(ui), hash)


def _usercachedir(ui, name=longname):
    '''Return the location of the "global" largefiles cache.'''
    path = ui.configpath(name, b'usercache')
    if path:
        return path

    hint = None

    if pycompat.iswindows:
        appdata = encoding.environ.get(
            b'LOCALAPPDATA', encoding.environ.get(b'APPDATA')
        )
        if appdata:
            return os.path.join(appdata, name)

        hint = _(b"define %s or %s in the environment, or set %s.usercache") % (
            b"LOCALAPPDATA",
            b"APPDATA",
            name,
        )
    elif pycompat.isdarwin:
        home = encoding.environ.get(b'HOME')
        if home:
            return os.path.join(home, b'Library', b'Caches', name)

        hint = _(b"define %s in the environment, or set %s.usercache") % (
            b"HOME",
            name,
        )
    elif pycompat.isposix:
        path = encoding.environ.get(b'XDG_CACHE_HOME')
        if path:
            return os.path.join(path, name)
        home = encoding.environ.get(b'HOME')
        if home:
            return os.path.join(home, b'.cache', name)

        hint = _(b"define %s or %s in the environment, or set %s.usercache") % (
            b"XDG_CACHE_HOME",
            b"HOME",
            name,
        )
    else:
        raise error.Abort(
            _(b'unknown operating system: %s\n') % pycompat.osname
        )

    raise error.Abort(_(b'unknown %s usercache location') % name, hint=hint)


def inusercache(ui, hash):
    path = usercachepath(ui, hash)
    return os.path.exists(path)


def findfile(repo, hash):
    """Return store path of the largefile with the specified hash.
    As a side effect, the file might be linked from user cache.
    Return None if the file can't be found locally."""
    path, exists = findstorepath(repo, hash)
    if exists:
        repo.ui.note(_(b'found %s in store\n') % hash)
        return path
    elif inusercache(repo.ui, hash):
        repo.ui.note(_(b'found %s in system cache\n') % hash)
        path = storepath(repo, hash)
        link(usercachepath(repo.ui, hash), path)
        return path
    return None


class largefilesdirstate(dirstate.dirstate):
    def __getitem__(self, key):
        return super(largefilesdirstate, self).__getitem__(unixpath(key))

    def set_tracked(self, f):
        return super(largefilesdirstate, self).set_tracked(unixpath(f))

    def set_untracked(self, f):
        return super(largefilesdirstate, self).set_untracked(unixpath(f))

    def normal(self, f, parentfiledata=None):
        # not sure if we should pass the `parentfiledata` down or throw it
        # away. So throwing it away to stay on the safe side.
        return super(largefilesdirstate, self).normal(unixpath(f))

    def remove(self, f):
        return super(largefilesdirstate, self).remove(unixpath(f))

    def add(self, f):
        return super(largefilesdirstate, self).add(unixpath(f))

    def drop(self, f):
        return super(largefilesdirstate, self).drop(unixpath(f))

    def forget(self, f):
        return super(largefilesdirstate, self).forget(unixpath(f))

    def normallookup(self, f):
        return super(largefilesdirstate, self).normallookup(unixpath(f))

    def _ignore(self, f):
        return False

    def write(self, tr=False):
        # (1) disable PENDING mode always
        #     (lfdirstate isn't yet managed as a part of the transaction)
        # (2) avoid develwarn 'use dirstate.write with ....'
        super(largefilesdirstate, self).write(None)


def openlfdirstate(ui, repo, create=True):
    """
    Return a dirstate object that tracks largefiles: i.e. its root is
    the repo root, but it is saved in .hg/largefiles/dirstate.
    """
    vfs = repo.vfs
    lfstoredir = longname
    opener = vfsmod.vfs(vfs.join(lfstoredir))
    use_dirstate_v2 = requirements.DIRSTATE_V2_REQUIREMENT in repo.requirements
    lfdirstate = largefilesdirstate(
        opener,
        ui,
        repo.root,
        repo.dirstate._validate,
        lambda: sparse.matcher(repo),
        repo.nodeconstants,
        use_dirstate_v2,
    )

    # If the largefiles dirstate does not exist, populate and create
    # it. This ensures that we create it on the first meaningful
    # largefiles operation in a new clone.
    if create and not vfs.exists(vfs.join(lfstoredir, b'dirstate')):
        matcher = getstandinmatcher(repo)
        standins = repo.dirstate.walk(
            matcher, subrepos=[], unknown=False, ignored=False
        )

        if len(standins) > 0:
            vfs.makedirs(lfstoredir)

        with lfdirstate.parentchange():
            for standin in standins:
                lfile = splitstandin(standin)
                lfdirstate.update_file(
                    lfile, p1_tracked=True, wc_tracked=True, possibly_dirty=True
                )
    return lfdirstate


def lfdirstatestatus(lfdirstate, repo):
    pctx = repo[b'.']
    match = matchmod.always()
    unsure, s = lfdirstate.status(
        match, subrepos=[], ignored=False, clean=False, unknown=False
    )
    modified, clean = s.modified, s.clean
    for lfile in unsure:
        try:
            fctx = pctx[standin(lfile)]
        except LookupError:
            fctx = None
        if not fctx or readasstandin(fctx) != hashfile(repo.wjoin(lfile)):
            modified.append(lfile)
        else:
            clean.append(lfile)
            lfdirstate.set_clean(lfile)
    return s


def listlfiles(repo, rev=None, matcher=None):
    """return a list of largefiles in the working copy or the
    specified changeset"""

    if matcher is None:
        matcher = getstandinmatcher(repo)

    # ignore unknown files in working directory
    return [
        splitstandin(f)
        for f in repo[rev].walk(matcher)
        if rev is not None or repo.dirstate[f] != b'?'
    ]


def instore(repo, hash, forcelocal=False):
    '''Return true if a largefile with the given hash exists in the store'''
    return os.path.exists(storepath(repo, hash, forcelocal))


def storepath(repo, hash, forcelocal=False):
    """Return the correct location in the repository largefiles store for a
    file with the given hash."""
    if not forcelocal and repo.shared():
        return repo.vfs.reljoin(repo.sharedpath, longname, hash)
    return repo.vfs.join(longname, hash)


def findstorepath(repo, hash):
    """Search through the local store path(s) to find the file for the given
    hash.  If the file is not found, its path in the primary store is returned.
    The return value is a tuple of (path, exists(path)).
    """
    # For shared repos, the primary store is in the share source.  But for
    # backward compatibility, force a lookup in the local store if it wasn't
    # found in the share source.
    path = storepath(repo, hash, False)

    if instore(repo, hash):
        return (path, True)
    elif repo.shared() and instore(repo, hash, True):
        return storepath(repo, hash, True), True

    return (path, False)


def copyfromcache(repo, hash, filename):
    """Copy the specified largefile from the repo or system cache to
    filename in the repository. Return true on success or false if the
    file was not found in either cache (which should not happened:
    this is meant to be called only after ensuring that the needed
    largefile exists in the cache)."""
    wvfs = repo.wvfs
    path = findfile(repo, hash)
    if path is None:
        return False
    wvfs.makedirs(wvfs.dirname(wvfs.join(filename)))
    # The write may fail before the file is fully written, but we
    # don't use atomic writes in the working copy.
    with open(path, b'rb') as srcfd, wvfs(filename, b'wb') as destfd:
        gothash = copyandhash(util.filechunkiter(srcfd), destfd)
    if gothash != hash:
        repo.ui.warn(
            _(b'%s: data corruption in %s with hash %s\n')
            % (filename, path, gothash)
        )
        wvfs.unlink(filename)
        return False
    return True


def copytostore(repo, ctx, file, fstandin):
    wvfs = repo.wvfs
    hash = readasstandin(ctx[fstandin])
    if instore(repo, hash):
        return
    if wvfs.exists(file):
        copytostoreabsolute(repo, wvfs.join(file), hash)
    else:
        repo.ui.warn(
            _(b"%s: largefile %s not available from local store\n")
            % (file, hash)
        )


def copyalltostore(repo, node):
    '''Copy all largefiles in a given revision to the store'''

    ctx = repo[node]
    for filename in ctx.files():
        realfile = splitstandin(filename)
        if realfile is not None and filename in ctx.manifest():
            copytostore(repo, ctx, realfile, filename)


def copytostoreabsolute(repo, file, hash):
    if inusercache(repo.ui, hash):
        link(usercachepath(repo.ui, hash), storepath(repo, hash))
    else:
        util.makedirs(os.path.dirname(storepath(repo, hash)))
        with open(file, b'rb') as srcf:
            with util.atomictempfile(
                storepath(repo, hash), createmode=repo.store.createmode
            ) as dstf:
                for chunk in util.filechunkiter(srcf):
                    dstf.write(chunk)
        linktousercache(repo, hash)


def linktousercache(repo, hash):
    """Link / copy the largefile with the specified hash from the store
    to the cache."""
    path = usercachepath(repo.ui, hash)
    link(storepath(repo, hash), path)


def getstandinmatcher(repo, rmatcher=None):
    '''Return a match object that applies rmatcher to the standin directory'''
    wvfs = repo.wvfs
    standindir = shortname

    # no warnings about missing files or directories
    badfn = lambda f, msg: None

    if rmatcher and not rmatcher.always():
        pats = [wvfs.join(standindir, pat) for pat in rmatcher.files()]
        if not pats:
            pats = [wvfs.join(standindir)]
        match = scmutil.match(repo[None], pats, badfn=badfn)
    else:
        # no patterns: relative to repo root
        match = scmutil.match(repo[None], [wvfs.join(standindir)], badfn=badfn)
    return match


def composestandinmatcher(repo, rmatcher):
    """Return a matcher that accepts standins corresponding to the
    files accepted by rmatcher. Pass the list of files in the matcher
    as the paths specified by the user."""
    smatcher = getstandinmatcher(repo, rmatcher)
    isstandin = smatcher.matchfn

    def composedmatchfn(f):
        return isstandin(f) and rmatcher.matchfn(splitstandin(f))

    smatcher.matchfn = composedmatchfn

    return smatcher


def standin(filename):
    """Return the repo-relative path to the standin for the specified big
    file."""
    # Notes:
    # 1) Some callers want an absolute path, but for instance addlargefiles
    #    needs it repo-relative so it can be passed to repo[None].add().  So
    #    leave it up to the caller to use repo.wjoin() to get an absolute path.
    # 2) Join with '/' because that's what dirstate always uses, even on
    #    Windows. Change existing separator to '/' first in case we are
    #    passed filenames from an external source (like the command line).
    return shortnameslash + util.pconvert(filename)


def isstandin(filename):
    """Return true if filename is a big file standin. filename must be
    in Mercurial's internal form (slash-separated)."""
    return filename.startswith(shortnameslash)


def splitstandin(filename):
    # Split on / because that's what dirstate always uses, even on Windows.
    # Change local separator to / first just in case we are passed filenames
    # from an external source (like the command line).
    bits = util.pconvert(filename).split(b'/', 1)
    if len(bits) == 2 and bits[0] == shortname:
        return bits[1]
    else:
        return None


def updatestandin(repo, lfile, standin):
    """Re-calculate hash value of lfile and write it into standin

    This assumes that "lfutil.standin(lfile) == standin", for efficiency.
    """
    file = repo.wjoin(lfile)
    if repo.wvfs.exists(lfile):
        hash = hashfile(file)
        executable = getexecutable(file)
        writestandin(repo, standin, hash, executable)
    else:
        raise error.Abort(_(b'%s: file not found!') % lfile)


def readasstandin(fctx):
    """read hex hash from given filectx of standin file

    This encapsulates how "standin" data is stored into storage layer."""
    return fctx.data().strip()


def writestandin(repo, standin, hash, executable):
    '''write hash to <repo.root>/<standin>'''
    repo.wwrite(standin, hash + b'\n', executable and b'x' or b'')


def copyandhash(instream, outfile):
    """Read bytes from instream (iterable) and write them to outfile,
    computing the SHA-1 hash of the data along the way. Return the hash."""
    hasher = hashutil.sha1(b'')
    for data in instream:
        hasher.update(data)
        outfile.write(data)
    return hex(hasher.digest())


def hashfile(file):
    if not os.path.exists(file):
        return b''
    with open(file, b'rb') as fd:
        return hexsha1(fd)


def getexecutable(filename):
    mode = os.stat(filename).st_mode
    return (
        (mode & stat.S_IXUSR)
        and (mode & stat.S_IXGRP)
        and (mode & stat.S_IXOTH)
    )


def urljoin(first, second, *arg):
    def join(left, right):
        if not left.endswith(b'/'):
            left += b'/'
        if right.startswith(b'/'):
            right = right[1:]
        return left + right

    url = join(first, second)
    for a in arg:
        url = join(url, a)
    return url


def hexsha1(fileobj):
    """hexsha1 returns the hex-encoded sha1 sum of the data in the file-like
    object data"""
    h = hashutil.sha1()
    for chunk in util.filechunkiter(fileobj):
        h.update(chunk)
    return hex(h.digest())


def httpsendfile(ui, filename):
    return httpconnection.httpsendfile(ui, filename, b'rb')


def unixpath(path):
    '''Return a version of path normalized for use with the lfdirstate.'''
    return util.pconvert(os.path.normpath(path))


def islfilesrepo(repo):
    '''Return true if the repo is a largefile repo.'''
    if b'largefiles' in repo.requirements and any(
        shortnameslash in f[1] for f in repo.store.datafiles()
    ):
        return True

    return any(openlfdirstate(repo.ui, repo, False))


class storeprotonotcapable(Exception):
    def __init__(self, storetypes):
        self.storetypes = storetypes


def getstandinsstate(repo):
    standins = []
    matcher = getstandinmatcher(repo)
    wctx = repo[None]
    for standin in repo.dirstate.walk(
        matcher, subrepos=[], unknown=False, ignored=False
    ):
        lfile = splitstandin(standin)
        try:
            hash = readasstandin(wctx[standin])
        except IOError:
            hash = None
        standins.append((lfile, hash))
    return standins


def synclfdirstate(repo, lfdirstate, lfile, normallookup):
    lfstandin = standin(lfile)
    if lfstandin not in repo.dirstate:
        lfdirstate.update_file(lfile, p1_tracked=False, wc_tracked=False)
    else:
        stat = repo.dirstate._map[lfstandin]
        state, mtime = stat.state, stat.mtime
        if state == b'n':
            if normallookup or mtime < 0 or not repo.wvfs.exists(lfile):
                # state 'n' doesn't ensure 'clean' in this case
                lfdirstate.update_file(
                    lfile, p1_tracked=True, wc_tracked=True, possibly_dirty=True
                )
            else:
                lfdirstate.update_file(lfile, p1_tracked=True, wc_tracked=True)
        elif state == b'm':
            lfdirstate.update_file(
                lfile, p1_tracked=True, wc_tracked=True, merged=True
            )
        elif state == b'r':
            lfdirstate.update_file(lfile, p1_tracked=True, wc_tracked=False)
        elif state == b'a':
            lfdirstate.update_file(lfile, p1_tracked=False, wc_tracked=True)


def markcommitted(orig, ctx, node):
    repo = ctx.repo()

    lfdirstate = openlfdirstate(repo.ui, repo)
    with lfdirstate.parentchange():
        orig(node)

        # ATTENTION: "ctx.files()" may differ from "repo[node].files()"
        # because files coming from the 2nd parent are omitted in the latter.
        #
        # The former should be used to get targets of "synclfdirstate",
        # because such files:
        # - are marked as "a" by "patch.patch()" (e.g. via transplant), and
        # - have to be marked as "n" after commit, but
        # - aren't listed in "repo[node].files()"

        for f in ctx.files():
            lfile = splitstandin(f)
            if lfile is not None:
                synclfdirstate(repo, lfdirstate, lfile, False)
    lfdirstate.write()

    # As part of committing, copy all of the largefiles into the cache.
    #
    # Using "node" instead of "ctx" implies additional "repo[node]"
    # lookup while copyalltostore(), but can omit redundant check for
    # files comming from the 2nd parent, which should exist in store
    # at merging.
    copyalltostore(repo, node)


def getlfilestoupdate(oldstandins, newstandins):
    changedstandins = set(oldstandins).symmetric_difference(set(newstandins))
    filelist = []
    for f in changedstandins:
        if f[0] not in filelist:
            filelist.append(f[0])
    return filelist


def getlfilestoupload(repo, missing, addfunc):
    makeprogress = repo.ui.makeprogress
    with makeprogress(
        _(b'finding outgoing largefiles'),
        unit=_(b'revisions'),
        total=len(missing),
    ) as progress:
        for i, n in enumerate(missing):
            progress.update(i)
            parents = [p for p in repo[n].parents() if p != repo.nullid]

            with lfstatus(repo, value=False):
                ctx = repo[n]

            files = set(ctx.files())
            if len(parents) == 2:
                mc = ctx.manifest()
                mp1 = ctx.p1().manifest()
                mp2 = ctx.p2().manifest()
                for f in mp1:
                    if f not in mc:
                        files.add(f)
                for f in mp2:
                    if f not in mc:
                        files.add(f)
                for f in mc:
                    if mc[f] != mp1.get(f, None) or mc[f] != mp2.get(f, None):
                        files.add(f)
            for fn in files:
                if isstandin(fn) and fn in ctx:
                    addfunc(fn, readasstandin(ctx[fn]))


def updatestandinsbymatch(repo, match):
    """Update standins in the working directory according to specified match

    This returns (possibly modified) ``match`` object to be used for
    subsequent commit process.
    """

    ui = repo.ui

    # Case 1: user calls commit with no specific files or
    # include/exclude patterns: refresh and commit all files that
    # are "dirty".
    if match is None or match.always():
        # Spend a bit of time here to get a list of files we know
        # are modified so we can compare only against those.
        # It can cost a lot of time (several seconds)
        # otherwise to update all standins if the largefiles are
        # large.
        lfdirstate = openlfdirstate(ui, repo)
        dirtymatch = matchmod.always()
        unsure, s = lfdirstate.status(
            dirtymatch, subrepos=[], ignored=False, clean=False, unknown=False
        )
        modifiedfiles = unsure + s.modified + s.added + s.removed
        lfiles = listlfiles(repo)
        # this only loops through largefiles that exist (not
        # removed/renamed)
        for lfile in lfiles:
            if lfile in modifiedfiles:
                fstandin = standin(lfile)
                if repo.wvfs.exists(fstandin):
                    # this handles the case where a rebase is being
                    # performed and the working copy is not updated
                    # yet.
                    if repo.wvfs.exists(lfile):
                        updatestandin(repo, lfile, fstandin)

        return match

    lfiles = listlfiles(repo)
    match._files = repo._subdirlfs(match.files(), lfiles)

    # Case 2: user calls commit with specified patterns: refresh
    # any matching big files.
    smatcher = composestandinmatcher(repo, match)
    standins = repo.dirstate.walk(
        smatcher, subrepos=[], unknown=False, ignored=False
    )

    # No matching big files: get out of the way and pass control to
    # the usual commit() method.
    if not standins:
        return match

    # Refresh all matching big files.  It's possible that the
    # commit will end up failing, in which case the big files will
    # stay refreshed.  No harm done: the user modified them and
    # asked to commit them, so sooner or later we're going to
    # refresh the standins.  Might as well leave them refreshed.
    lfdirstate = openlfdirstate(ui, repo)
    for fstandin in standins:
        lfile = splitstandin(fstandin)
        if lfdirstate[lfile] != b'r':
            updatestandin(repo, lfile, fstandin)

    # Cook up a new matcher that only matches regular files or
    # standins corresponding to the big files requested by the
    # user.  Have to modify _files to prevent commit() from
    # complaining "not tracked" for big files.
    match = copy.copy(match)
    origmatchfn = match.matchfn

    # Check both the list of largefiles and the list of
    # standins because if a largefile was removed, it
    # won't be in the list of largefiles at this point
    match._files += sorted(standins)

    actualfiles = []
    for f in match._files:
        fstandin = standin(f)

        # For largefiles, only one of the normal and standin should be
        # committed (except if one of them is a remove).  In the case of a
        # standin removal, drop the normal file if it is unknown to dirstate.
        # Thus, skip plain largefile names but keep the standin.
        if f in lfiles or fstandin in standins:
            if repo.dirstate[fstandin] != b'r':
                if repo.dirstate[f] != b'r':
                    continue
            elif repo.dirstate[f] == b'?':
                continue

        actualfiles.append(f)
    match._files = actualfiles

    def matchfn(f):
        if origmatchfn(f):
            return f not in lfiles
        else:
            return f in standins

    match.matchfn = matchfn

    return match


class automatedcommithook(object):
    """Stateful hook to update standins at the 1st commit of resuming

    For efficiency, updating standins in the working directory should
    be avoided while automated committing (like rebase, transplant and
    so on), because they should be updated before committing.

    But the 1st commit of resuming automated committing (e.g. ``rebase
    --continue``) should update them, because largefiles may be
    modified manually.
    """

    def __init__(self, resuming):
        self.resuming = resuming

    def __call__(self, repo, match):
        if self.resuming:
            self.resuming = False  # avoids updating at subsequent commits
            return updatestandinsbymatch(repo, match)
        else:
            return match


def getstatuswriter(ui, repo, forcibly=None):
    """Return the function to write largefiles specific status out

    If ``forcibly`` is ``None``, this returns the last element of
    ``repo._lfstatuswriters`` as "default" writer function.

    Otherwise, this returns the function to always write out (or
    ignore if ``not forcibly``) status.
    """
    if forcibly is None and util.safehasattr(repo, b'_largefilesenabled'):
        return repo._lfstatuswriters[-1]
    else:
        if forcibly:
            return ui.status  # forcibly WRITE OUT
        else:
            return lambda *msg, **opts: None  # forcibly IGNORE



# extensions.py - extension handling for mercurial
#
# Copyright 2005-2007 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import ast
import collections
import functools
import imp
import inspect
import os

from .i18n import (
    _,
    gettext,
)
from .pycompat import (
    getattr,
    open,
    setattr,
)

from . import (
    cmdutil,
    configitems,
    error,
    pycompat,
    util,
)

from .utils import stringutil

_extensions = {}
_disabledextensions = {}
_aftercallbacks = {}
_order = []
_builtin = {
    b'hbisect',
    b'bookmarks',
    b'color',
    b'parentrevspec',
    b'progress',
    b'interhg',
    b'inotify',
    b'hgcia',
    b'shelve',
}


def extensions(ui=None):
    if ui:

        def enabled(name):
            for format in [b'%s', b'hgext.%s']:
                conf = ui.config(b'extensions', format % name)
                if conf is not None and not conf.startswith(b'!'):
                    return True

    else:
        enabled = lambda name: True
    for name in _order:
        module = _extensions[name]
        if module and enabled(name):
            yield name, module


def find(name):
    '''return module with given extension name'''
    mod = None
    try:
        mod = _extensions[name]
    except KeyError:
        for k, v in pycompat.iteritems(_extensions):
            if k.endswith(b'.' + name) or k.endswith(b'/' + name):
                mod = v
                break
    if not mod:
        raise KeyError(name)
    return mod


def loadpath(path, module_name):
    module_name = module_name.replace(b'.', b'_')
    path = util.normpath(util.expandpath(path))
    module_name = pycompat.fsdecode(module_name)
    path = pycompat.fsdecode(path)
    if os.path.isdir(path):
        # module/__init__.py style
        d, f = os.path.split(path)
        fd, fpath, desc = imp.find_module(f, [d])
        # When https://github.com/python/typeshed/issues/3466 is fixed
        # and in a pytype release we can drop this disable.
        return imp.load_module(
            module_name, fd, fpath, desc  # pytype: disable=wrong-arg-types
        )
    else:
        try:
            return imp.load_source(module_name, path)
        except IOError as exc:
            if not exc.filename:
                exc.filename = path  # python does not fill this
            raise


def _importh(name):
    """import and return the <name> module"""
    mod = __import__(pycompat.sysstr(name))
    components = name.split(b'.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def _importext(name, path=None, reportfunc=None):
    if path:
        # the module will be loaded in sys.modules
        # choose an unique name so that it doesn't
        # conflicts with other modules
        mod = loadpath(path, b'hgext.%s' % name)
    else:
        try:
            mod = _importh(b"hgext.%s" % name)
        except ImportError as err:
            if reportfunc:
                reportfunc(err, b"hgext.%s" % name, b"hgext3rd.%s" % name)
            try:
                mod = _importh(b"hgext3rd.%s" % name)
            except ImportError as err:
                if reportfunc:
                    reportfunc(err, b"hgext3rd.%s" % name, name)
                mod = _importh(name)
    return mod


def _reportimporterror(ui, err, failed, next):
    # note: this ui.log happens before --debug is processed,
    #       Use --config ui.debug=1 to see them.
    ui.log(
        b'extension',
        b'    - could not import %s (%s): trying %s\n',
        failed,
        stringutil.forcebytestr(err),
        next,
    )
    if ui.debugflag and ui.configbool(b'devel', b'debug.extensions'):
        ui.traceback()


def _rejectunicode(name, xs):
    if isinstance(xs, (list, set, tuple)):
        for x in xs:
            _rejectunicode(name, x)
    elif isinstance(xs, dict):
        for k, v in xs.items():
            _rejectunicode(name, k)
            _rejectunicode(b'%s.%s' % (name, stringutil.forcebytestr(k)), v)
    elif isinstance(xs, type(u'')):
        raise error.ProgrammingError(
            b"unicode %r found in %s" % (xs, name),
            hint=b"use b'' to make it byte string",
        )


# attributes set by registrar.command
_cmdfuncattrs = (b'norepo', b'optionalrepo', b'inferrepo')


def _validatecmdtable(ui, cmdtable):
    """Check if extension commands have required attributes"""
    for c, e in pycompat.iteritems(cmdtable):
        f = e[0]
        missing = [a for a in _cmdfuncattrs if not util.safehasattr(f, a)]
        if not missing:
            continue
        raise error.ProgrammingError(
            b'missing attributes: %s' % b', '.join(missing),
            hint=b"use @command decorator to register '%s'" % c,
        )


def _validatetables(ui, mod):
    """Sanity check for loadable tables provided by extension module"""
    for t in [b'cmdtable', b'colortable', b'configtable']:
        _rejectunicode(t, getattr(mod, t, {}))
    for t in [
        b'filesetpredicate',
        b'internalmerge',
        b'revsetpredicate',
        b'templatefilter',
        b'templatefunc',
        b'templatekeyword',
    ]:
        o = getattr(mod, t, None)
        if o:
            _rejectunicode(t, o._table)
    _validatecmdtable(ui, getattr(mod, 'cmdtable', {}))


def load(ui, name, path, loadingtime=None):
    if name.startswith(b'hgext.') or name.startswith(b'hgext/'):
        shortname = name[6:]
    else:
        shortname = name
    if shortname in _builtin:
        return None
    if shortname in _extensions:
        return _extensions[shortname]
    ui.log(b'extension', b'  - loading extension: %s\n', shortname)
    _extensions[shortname] = None
    with util.timedcm('load extension %s', shortname) as stats:
        mod = _importext(name, path, bind(_reportimporterror, ui))
    ui.log(b'extension', b'  > %s extension loaded in %s\n', shortname, stats)
    if loadingtime is not None:
        loadingtime[shortname] += stats.elapsed

    # Before we do anything with the extension, check against minimum stated
    # compatibility. This gives extension authors a mechanism to have their
    # extensions short circuit when loaded with a known incompatible version
    # of Mercurial.
    minver = getattr(mod, 'minimumhgversion', None)
    if minver:
        curver = util.versiontuple(n=2)

        if None in curver or util.versiontuple(minver, 2) > curver:
            msg = _(
                b'(third party extension %s requires version %s or newer '
                b'of Mercurial (current: %s); disabling)\n'
            )
            ui.warn(msg % (shortname, minver, util.version()))
            return
    ui.log(b'extension', b'    - validating extension tables: %s\n', shortname)
    _validatetables(ui, mod)

    _extensions[shortname] = mod
    _order.append(shortname)
    ui.log(
        b'extension', b'    - invoking registered callbacks: %s\n', shortname
    )
    with util.timedcm('callbacks extension %s', shortname) as stats:
        for fn in _aftercallbacks.get(shortname, []):
            fn(loaded=True)
    ui.log(b'extension', b'    > callbacks completed in %s\n', stats)
    return mod


def _runuisetup(name, ui):
    uisetup = getattr(_extensions[name], 'uisetup', None)
    if uisetup:
        try:
            uisetup(ui)
        except Exception as inst:
            ui.traceback(force=True)
            msg = stringutil.forcebytestr(inst)
            ui.warn(_(b"*** failed to set up extension %s: %s\n") % (name, msg))
            return False
    return True


def _runextsetup(name, ui):
    extsetup = getattr(_extensions[name], 'extsetup', None)
    if extsetup:
        try:
            extsetup(ui)
        except Exception as inst:
            ui.traceback(force=True)
            msg = stringutil.forcebytestr(inst)
            ui.warn(_(b"*** failed to set up extension %s: %s\n") % (name, msg))
            return False
    return True


def loadall(ui, whitelist=None):
    loadingtime = collections.defaultdict(int)
    result = ui.configitems(b"extensions")
    if whitelist is not None:
        result = [(k, v) for (k, v) in result if k in whitelist]
    newindex = len(_order)
    ui.log(
        b'extension',
        b'loading %sextensions\n',
        b'additional ' if newindex else b'',
    )
    ui.log(b'extension', b'- processing %d entries\n', len(result))
    with util.timedcm('load all extensions') as stats:
        for (name, path) in result:
            if path:
                if path[0:1] == b'!':
                    if name not in _disabledextensions:
                        ui.log(
                            b'extension',
                            b'  - skipping disabled extension: %s\n',
                            name,
                        )
                    _disabledextensions[name] = path[1:]
                    continue
            try:
                load(ui, name, path, loadingtime)
            except Exception as inst:
                msg = stringutil.forcebytestr(inst)
                if path:
                    ui.warn(
                        _(b"*** failed to import extension %s from %s: %s\n")
                        % (name, path, msg)
                    )
                else:
                    ui.warn(
                        _(b"*** failed to import extension %s: %s\n")
                        % (name, msg)
                    )
                if isinstance(inst, error.Hint) and inst.hint:
                    ui.warn(_(b"*** (%s)\n") % inst.hint)
                ui.traceback()

    ui.log(
        b'extension',
        b'> loaded %d extensions, total time %s\n',
        len(_order) - newindex,
        stats,
    )
    # list of (objname, loadermod, loadername) tuple:
    # - objname is the name of an object in extension module,
    #   from which extra information is loaded
    # - loadermod is the module where loader is placed
    # - loadername is the name of the function,
    #   which takes (ui, extensionname, extraobj) arguments
    #
    # This one is for the list of item that must be run before running any setup
    earlyextraloaders = [
        (b'configtable', configitems, b'loadconfigtable'),
    ]

    ui.log(b'extension', b'- loading configtable attributes\n')
    _loadextra(ui, newindex, earlyextraloaders)

    broken = set()
    ui.log(b'extension', b'- executing uisetup hooks\n')
    with util.timedcm('all uisetup') as alluisetupstats:
        for name in _order[newindex:]:
            ui.log(b'extension', b'  - running uisetup for %s\n', name)
            with util.timedcm('uisetup %s', name) as stats:
                if not _runuisetup(name, ui):
                    ui.log(
                        b'extension',
                        b'    - the %s extension uisetup failed\n',
                        name,
                    )
                    broken.add(name)
            ui.log(b'extension', b'  > uisetup for %s took %s\n', name, stats)
            loadingtime[name] += stats.elapsed
    ui.log(b'extension', b'> all uisetup took %s\n', alluisetupstats)

    ui.log(b'extension', b'- executing extsetup hooks\n')
    with util.timedcm('all extsetup') as allextetupstats:
        for name in _order[newindex:]:
            if name in broken:
                continue
            ui.log(b'extension', b'  - running extsetup for %s\n', name)
            with util.timedcm('extsetup %s', name) as stats:
                if not _runextsetup(name, ui):
                    ui.log(
                        b'extension',
                        b'    - the %s extension extsetup failed\n',
                        name,
                    )
                    broken.add(name)
            ui.log(b'extension', b'  > extsetup for %s took %s\n', name, stats)
            loadingtime[name] += stats.elapsed
    ui.log(b'extension', b'> all extsetup took %s\n', allextetupstats)

    for name in broken:
        ui.log(b'extension', b'    - disabling broken %s extension\n', name)
        _extensions[name] = None

    # Call aftercallbacks that were never met.
    ui.log(b'extension', b'- executing remaining aftercallbacks\n')
    with util.timedcm('aftercallbacks') as stats:
        for shortname in _aftercallbacks:
            if shortname in _extensions:
                continue

            for fn in _aftercallbacks[shortname]:
                ui.log(
                    b'extension',
                    b'  - extension %s not loaded, notify callbacks\n',
                    shortname,
                )
                fn(loaded=False)
    ui.log(b'extension', b'> remaining aftercallbacks completed in %s\n', stats)

    # loadall() is called multiple times and lingering _aftercallbacks
    # entries could result in double execution. See issue4646.
    _aftercallbacks.clear()

    # delay importing avoids cyclic dependency (especially commands)
    from . import (
        color,
        commands,
        filemerge,
        fileset,
        revset,
        templatefilters,
        templatefuncs,
        templatekw,
    )

    # list of (objname, loadermod, loadername) tuple:
    # - objname is the name of an object in extension module,
    #   from which extra information is loaded
    # - loadermod is the module where loader is placed
    # - loadername is the name of the function,
    #   which takes (ui, extensionname, extraobj) arguments
    ui.log(b'extension', b'- loading extension registration objects\n')
    extraloaders = [
        (b'cmdtable', commands, b'loadcmdtable'),
        (b'colortable', color, b'loadcolortable'),
        (b'filesetpredicate', fileset, b'loadpredicate'),
        (b'internalmerge', filemerge, b'loadinternalmerge'),
        (b'revsetpredicate', revset, b'loadpredicate'),
        (b'templatefilter', templatefilters, b'loadfilter'),
        (b'templatefunc', templatefuncs, b'loadfunction'),
        (b'templatekeyword', templatekw, b'loadkeyword'),
    ]
    with util.timedcm('load registration objects') as stats:
        _loadextra(ui, newindex, extraloaders)
    ui.log(
        b'extension',
        b'> extension registration object loading took %s\n',
        stats,
    )

    # Report per extension loading time (except reposetup)
    for name in sorted(loadingtime):
        ui.log(
            b'extension',
            b'> extension %s take a total of %s to load\n',
            name,
            util.timecount(loadingtime[name]),
        )

    ui.log(b'extension', b'extension loading complete\n')


def _loadextra(ui, newindex, extraloaders):
    for name in _order[newindex:]:
        module = _extensions[name]
        if not module:
            continue  # loading this module failed

        for objname, loadermod, loadername in extraloaders:
            extraobj = getattr(module, objname, None)
            if extraobj is not None:
                getattr(loadermod, loadername)(ui, name, extraobj)


def afterloaded(extension, callback):
    """Run the specified function after a named extension is loaded.

    If the named extension is already loaded, the callback will be called
    immediately.

    If the named extension never loads, the callback will be called after
    all extensions have been loaded.

    The callback receives the named argument ``loaded``, which is a boolean
    indicating whether the dependent extension actually loaded.
    """

    if extension in _extensions:
        # Report loaded as False if the extension is disabled
        loaded = _extensions[extension] is not None
        callback(loaded=loaded)
    else:
        _aftercallbacks.setdefault(extension, []).append(callback)


def populateui(ui):
    """Run extension hooks on the given ui to populate additional members,
    extend the class dynamically, etc.

    This will be called after the configuration is loaded, and/or extensions
    are loaded. In general, it's once per ui instance, but in command-server
    and hgweb, this may be called more than once with the same ui.
    """
    for name, mod in extensions(ui):
        hook = getattr(mod, 'uipopulate', None)
        if not hook:
            continue
        try:
            hook(ui)
        except Exception as inst:
            ui.traceback(force=True)
            ui.warn(
                _(b'*** failed to populate ui by extension %s: %s\n')
                % (name, stringutil.forcebytestr(inst))
            )


def bind(func, *args):
    """Partial function application

    Returns a new function that is the partial application of args and kwargs
    to func.  For example,

        f(1, 2, bar=3) === bind(f, 1)(2, bar=3)"""
    assert callable(func)

    def closure(*a, **kw):
        return func(*(args + a), **kw)

    return closure


def _updatewrapper(wrap, origfn, unboundwrapper):
    '''Copy and add some useful attributes to wrapper'''
    try:
        wrap.__name__ = origfn.__name__
    except AttributeError:
        pass
    wrap.__module__ = getattr(origfn, '__module__')
    wrap.__doc__ = getattr(origfn, '__doc__')
    wrap.__dict__.update(getattr(origfn, '__dict__', {}))
    wrap._origfunc = origfn
    wrap._unboundwrapper = unboundwrapper


def wrapcommand(table, command, wrapper, synopsis=None, docstring=None):
    '''Wrap the command named `command' in table

    Replace command in the command table with wrapper. The wrapped command will
    be inserted into the command table specified by the table argument.

    The wrapper will be called like

      wrapper(orig, *args, **kwargs)

    where orig is the original (wrapped) function, and *args, **kwargs
    are the arguments passed to it.

    Optionally append to the command synopsis and docstring, used for help.
    For example, if your extension wraps the ``bookmarks`` command to add the
    flags ``--remote`` and ``--all`` you might call this function like so:

      synopsis = ' [-a] [--remote]'
      docstring = """

      The ``remotenames`` extension adds the ``--remote`` and ``--all`` (``-a``)
      flags to the bookmarks command. Either flag will show the remote bookmarks
      known to the repository; ``--remote`` will also suppress the output of the
      local bookmarks.
      """

      extensions.wrapcommand(commands.table, 'bookmarks', exbookmarks,
                             synopsis, docstring)
    '''
    assert callable(wrapper)
    aliases, entry = cmdutil.findcmd(command, table)
    for alias, e in pycompat.iteritems(table):
        if e is entry:
            key = alias
            break

    origfn = entry[0]
    wrap = functools.partial(
        util.checksignature(wrapper), util.checksignature(origfn)
    )
    _updatewrapper(wrap, origfn, wrapper)
    if docstring is not None:
        wrap.__doc__ += docstring

    newentry = list(entry)
    newentry[0] = wrap
    if synopsis is not None:
        newentry[2] += synopsis
    table[key] = tuple(newentry)
    return entry


def wrapfilecache(cls, propname, wrapper):
    """Wraps a filecache property.

    These can't be wrapped using the normal wrapfunction.
    """
    propname = pycompat.sysstr(propname)
    assert callable(wrapper)
    for currcls in cls.__mro__:
        if propname in currcls.__dict__:
            origfn = currcls.__dict__[propname].func
            assert callable(origfn)

            def wrap(*args, **kwargs):
                return wrapper(origfn, *args, **kwargs)

            currcls.__dict__[propname].func = wrap
            break

    if currcls is object:
        raise AttributeError("type '%s' has no property '%s'" % (cls, propname))


class wrappedfunction(object):
    '''context manager for temporarily wrapping a function'''

    def __init__(self, container, funcname, wrapper):
        assert callable(wrapper)
        self._container = container
        self._funcname = funcname
        self._wrapper = wrapper

    def __enter__(self):
        wrapfunction(self._container, self._funcname, self._wrapper)

    def __exit__(self, exctype, excvalue, traceback):
        unwrapfunction(self._container, self._funcname, self._wrapper)


def wrapfunction(container, funcname, wrapper):
    """Wrap the function named funcname in container

    Replace the funcname member in the given container with the specified
    wrapper. The container is typically a module, class, or instance.

    The wrapper will be called like

      wrapper(orig, *args, **kwargs)

    where orig is the original (wrapped) function, and *args, **kwargs
    are the arguments passed to it.

    Wrapping methods of the repository object is not recommended since
    it conflicts with extensions that extend the repository by
    subclassing. All extensions that need to extend methods of
    localrepository should use this subclassing trick: namely,
    reposetup() should look like

      def reposetup(ui, repo):
          class myrepo(repo.__class__):
              def whatever(self, *args, **kwargs):
                  [...extension stuff...]
                  super(myrepo, self).whatever(*args, **kwargs)
                  [...extension stuff...]

          repo.__class__ = myrepo

    In general, combining wrapfunction() with subclassing does not
    work. Since you cannot control what other extensions are loaded by
    your end users, you should play nicely with others by using the
    subclass trick.
    """
    assert callable(wrapper)

    origfn = getattr(container, funcname)
    assert callable(origfn)
    if inspect.ismodule(container):
        # origfn is not an instance or class method. "partial" can be used.
        # "partial" won't insert a frame in traceback.
        wrap = functools.partial(wrapper, origfn)
    else:
        # "partial" cannot be safely used. Emulate its effect by using "bind".
        # The downside is one more frame in traceback.
        wrap = bind(wrapper, origfn)
    _updatewrapper(wrap, origfn, wrapper)
    setattr(container, funcname, wrap)
    return origfn


def unwrapfunction(container, funcname, wrapper=None):
    """undo wrapfunction

    If wrappers is None, undo the last wrap. Otherwise removes the wrapper
    from the chain of wrappers.

    Return the removed wrapper.
    Raise IndexError if wrapper is None and nothing to unwrap; ValueError if
    wrapper is not None but is not found in the wrapper chain.
    """
    chain = getwrapperchain(container, funcname)
    origfn = chain.pop()
    if wrapper is None:
        wrapper = chain[0]
    chain.remove(wrapper)
    setattr(container, funcname, origfn)
    for w in reversed(chain):
        wrapfunction(container, funcname, w)
    return wrapper


def getwrapperchain(container, funcname):
    """get a chain of wrappers of a function

    Return a list of functions: [newest wrapper, ..., oldest wrapper, origfunc]

    The wrapper functions are the ones passed to wrapfunction, whose first
    argument is origfunc.
    """
    result = []
    fn = getattr(container, funcname)
    while fn:
        assert callable(fn)
        result.append(getattr(fn, '_unboundwrapper', fn))
        fn = getattr(fn, '_origfunc', None)
    return result


def _disabledpaths():
    '''find paths of disabled extensions. returns a dict of {name: path}'''
    import hgext

    # The hgext might not have a __file__ attribute (e.g. in PyOxidizer) and
    # it might not be on a filesystem even if it does.
    if util.safehasattr(hgext, '__file__'):
        extpath = os.path.dirname(
            util.abspath(pycompat.fsencode(hgext.__file__))
        )
        try:
            files = os.listdir(extpath)
        except OSError:
            return {}
    else:
        return {}

    exts = {}
    for e in files:
        if e.endswith(b'.py'):
            name = e.rsplit(b'.', 1)[0]
            path = os.path.join(extpath, e)
        else:
            name = e
            path = os.path.join(extpath, e, b'__init__.py')
            if not os.path.exists(path):
                continue
        if name in exts or name in _order or name == b'__init__':
            continue
        exts[name] = path
    for name, path in pycompat.iteritems(_disabledextensions):
        # If no path was provided for a disabled extension (e.g. "color=!"),
        # don't replace the path we already found by the scan above.
        if path:
            exts[name] = path
    return exts


def _moduledoc(file):
    """return the top-level python documentation for the given file

    Loosely inspired by pydoc.source_synopsis(), but rewritten to
    handle triple quotes and to return the whole text instead of just
    the synopsis"""
    result = []

    line = file.readline()
    while line[:1] == b'#' or not line.strip():
        line = file.readline()
        if not line:
            break

    start = line[:3]
    if start == b'"""' or start == b"'''":
        line = line[3:]
        while line:
            if line.rstrip().endswith(start):
                line = line.split(start)[0]
                if line:
                    result.append(line)
                break
            elif not line:
                return None  # unmatched delimiter
            result.append(line)
            line = file.readline()
    else:
        return None

    return b''.join(result)


def _disabledhelp(path):
    '''retrieve help synopsis of a disabled extension (without importing)'''
    try:
        with open(path, b'rb') as src:
            doc = _moduledoc(src)
    except IOError:
        return

    if doc:  # extracting localized synopsis
        return gettext(doc)
    else:
        return _(b'(no help text available)')


def disabled():
    '''find disabled extensions from hgext. returns a dict of {name: desc}'''
    try:
        from hgext import __index__  # pytype: disable=import-error

        return {
            name: gettext(desc)
            for name, desc in pycompat.iteritems(__index__.docs)
            if name not in _order
        }
    except (ImportError, AttributeError):
        pass

    paths = _disabledpaths()
    if not paths:
        return {}

    exts = {}
    for name, path in pycompat.iteritems(paths):
        doc = _disabledhelp(path)
        if doc and name != b'__index__':
            exts[name] = doc.splitlines()[0]

    return exts


def disabled_help(name):
    """Obtain the full help text for a disabled extension, or None."""
    paths = _disabledpaths()
    if name in paths:
        return _disabledhelp(paths[name])


def _walkcommand(node):
    """Scan @command() decorators in the tree starting at node"""
    todo = collections.deque([node])
    while todo:
        node = todo.popleft()
        if not isinstance(node, ast.FunctionDef):
            todo.extend(ast.iter_child_nodes(node))
            continue
        for d in node.decorator_list:
            if not isinstance(d, ast.Call):
                continue
            if not isinstance(d.func, ast.Name):
                continue
            if d.func.id != 'command':
                continue
            yield d


def _disabledcmdtable(path):
    """Construct a dummy command table without loading the extension module

    This may raise IOError or SyntaxError.
    """
    with open(path, b'rb') as src:
        root = ast.parse(src.read(), path)
    cmdtable = {}
    for node in _walkcommand(root):
        if not node.args:
            continue
        a = node.args[0]
        if isinstance(a, ast.Str):
            name = pycompat.sysbytes(a.s)
        elif pycompat.ispy3 and isinstance(a, ast.Bytes):
            name = a.s
        else:
            continue
        cmdtable[name] = (None, [], b'')
    return cmdtable


def _finddisabledcmd(ui, cmd, name, path, strict):
    try:
        cmdtable = _disabledcmdtable(path)
    except (IOError, SyntaxError):
        return
    try:
        aliases, entry = cmdutil.findcmd(cmd, cmdtable, strict)
    except (error.AmbiguousCommand, error.UnknownCommand):
        return
    for c in aliases:
        if c.startswith(cmd):
            cmd = c
            break
    else:
        cmd = aliases[0]
    doc = _disabledhelp(path)
    return (cmd, name, doc)


def disabledcmd(ui, cmd, strict=False):
    """find cmd from disabled extensions without importing.
    returns (cmdname, extname, doc)"""

    paths = _disabledpaths()
    if not paths:
        raise error.UnknownCommand(cmd)

    ext = None
    # first, search for an extension with the same name as the command
    path = paths.pop(cmd, None)
    if path:
        ext = _finddisabledcmd(ui, cmd, cmd, path, strict=strict)
    if not ext:
        # otherwise, interrogate each extension until there's a match
        for name, path in pycompat.iteritems(paths):
            ext = _finddisabledcmd(ui, cmd, name, path, strict=strict)
            if ext:
                break
    if ext:
        return ext

    raise error.UnknownCommand(cmd)


def enabled(shortname=True):
    '''return a dict of {name: desc} of extensions'''
    exts = {}
    for ename, ext in extensions():
        doc = gettext(ext.__doc__) or _(b'(no help text available)')
        assert doc is not None  # help pytype
        if shortname:
            ename = ename.split(b'.')[-1]
        exts[ename] = doc.splitlines()[0].strip()

    return exts


def notloaded():
    '''return short names of extensions that failed to load'''
    return [
        name for name, mod in pycompat.iteritems(_extensions) if mod is None
    ]


def moduleversion(module):
    '''return version information from given module as a string'''
    if util.safehasattr(module, b'getversion') and callable(module.getversion):
        try:
            version = module.getversion()
        except Exception:
            version = b'unknown'

    elif util.safehasattr(module, b'__version__'):
        version = module.__version__
    else:
        version = b''
    if isinstance(version, (list, tuple)):
        version = b'.'.join(pycompat.bytestr(o) for o in version)
    else:
        # version data should be bytes, but not all extensions are ported
        # to py3.
        version = stringutil.forcebytestr(version)
    return version


def ismoduleinternal(module):
    exttestedwith = getattr(module, 'testedwith', None)
    return exttestedwith == b"ships-with-hg-core"



# closehead.py - Close arbitrary heads without checking them out first
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

'''close arbitrary heads without checking them out first'''

from __future__ import absolute_import

from mercurial.i18n import _
from mercurial import (
    bookmarks,
    cmdutil,
    context,
    error,
    pycompat,
    registrar,
    scmutil,
)

cmdtable = {}
command = registrar.command(cmdtable)
# Note for extension authors: ONLY specify testedwith = 'ships-with-hg-core' for
# extensions which SHIP WITH MERCURIAL. Non-mainline extensions should
# be specifying the version(s) of Mercurial they are tested with, or
# leave the attribute unspecified.
testedwith = b'ships-with-hg-core'

commitopts = cmdutil.commitopts
commitopts2 = cmdutil.commitopts2
commitopts3 = [(b'r', b'rev', [], _(b'revision to check'), _(b'REV'))]


@command(
    b'close-head|close-heads',
    commitopts + commitopts2 + commitopts3,
    _(b'[OPTION]... [REV]...'),
    helpcategory=command.CATEGORY_CHANGE_MANAGEMENT,
    inferrepo=True,
)
def close_branch(ui, repo, *revs, **opts):
    """close the given head revisions

    This is equivalent to checking out each revision in a clean tree and running
    ``hg commit --close-branch``, except that it doesn't change the working
    directory.

    The commit message must be specified with -l or -m.
    """

    def docommit(rev):
        cctx = context.memctx(
            repo,
            parents=[rev, None],
            text=message,
            files=[],
            filectxfn=None,
            user=opts.get(b'user'),
            date=opts.get(b'date'),
            extra=extra,
        )
        tr = repo.transaction(b'commit')
        ret = repo.commitctx(cctx, True)
        bookmarks.update(repo, [rev, None], ret)
        cctx.markcommitted(ret)
        tr.close()

    opts = pycompat.byteskwargs(opts)

    revs += tuple(opts.get(b'rev', []))
    revs = scmutil.revrange(repo, revs)

    if not revs:
        raise error.Abort(_(b'no revisions specified'))

    heads = []
    for branch in repo.branchmap():
        heads.extend(repo.branchheads(branch))
    heads = {repo[h].rev() for h in heads}
    for rev in revs:
        if rev not in heads:
            raise error.Abort(_(b'revision is not an open head: %d') % rev)

    message = cmdutil.logmessage(ui, opts)
    if not message:
        raise error.Abort(_(b"no commit message specified with -l or -m"))
    extra = {b'close': b'1'}

    with repo.wlock(), repo.lock():
        for rev in revs:
            r = repo[rev]
            branch = r.branch()
            extra[b'branch'] = branch
            docommit(r)
    return 0



# Copyright 2016-present Facebook. All Rights Reserved.
#
# commands: fastannotate commands
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import os

from mercurial.i18n import _
from mercurial import (
    commands,
    encoding,
    error,
    extensions,
    patch,
    pycompat,
    registrar,
    scmutil,
    util,
)

from . import (
    context as facontext,
    error as faerror,
    formatter as faformatter,
)

cmdtable = {}
command = registrar.command(cmdtable)


def _matchpaths(repo, rev, pats, opts, aopts=facontext.defaultopts):
    """generate paths matching given patterns"""
    perfhack = repo.ui.configbool(b'fastannotate', b'perfhack')

    # disable perfhack if:
    # a) any walkopt is used
    # b) if we treat pats as plain file names, some of them do not have
    #    corresponding linelog files
    if perfhack:
        # cwd related to reporoot
        reporoot = os.path.dirname(repo.path)
        reldir = os.path.relpath(encoding.getcwd(), reporoot)
        if reldir == b'.':
            reldir = b''
        if any(opts.get(o[1]) for o in commands.walkopts):  # a)
            perfhack = False
        else:  # b)
            relpats = [
                os.path.relpath(p, reporoot) if os.path.isabs(p) else p
                for p in pats
            ]
            # disable perfhack on '..' since it allows escaping from the repo
            if any(
                (
                    b'..' in f
                    or not os.path.isfile(
                        facontext.pathhelper(repo, f, aopts).linelogpath
                    )
                )
                for f in relpats
            ):
                perfhack = False

    # perfhack: emit paths directory without checking with manifest
    # this can be incorrect if the rev dos not have file.
    if perfhack:
        for p in relpats:
            yield os.path.join(reldir, p)
    else:

        def bad(x, y):
            raise error.Abort(b"%s: %s" % (x, y))

        ctx = scmutil.revsingle(repo, rev)
        m = scmutil.match(ctx, pats, opts, badfn=bad)
        for p in ctx.walk(m):
            yield p


fastannotatecommandargs = {
    'options': [
        (b'r', b'rev', b'.', _(b'annotate the specified revision'), _(b'REV')),
        (b'u', b'user', None, _(b'list the author (long with -v)')),
        (b'f', b'file', None, _(b'list the filename')),
        (b'd', b'date', None, _(b'list the date (short with -q)')),
        (b'n', b'number', None, _(b'list the revision number (default)')),
        (b'c', b'changeset', None, _(b'list the changeset')),
        (
            b'l',
            b'line-number',
            None,
            _(b'show line number at the first appearance'),
        ),
        (
            b'e',
            b'deleted',
            None,
            _(b'show deleted lines (slow) (EXPERIMENTAL)'),
        ),
        (
            b'',
            b'no-content',
            None,
            _(b'do not show file content (EXPERIMENTAL)'),
        ),
        (b'', b'no-follow', None, _(b"don't follow copies and renames")),
        (
            b'',
            b'linear',
            None,
            _(
                b'enforce linear history, ignore second parent '
                b'of merges (EXPERIMENTAL)'
            ),
        ),
        (
            b'',
            b'long-hash',
            None,
            _(b'show long changeset hash (EXPERIMENTAL)'),
        ),
        (
            b'',
            b'rebuild',
            None,
            _(b'rebuild cache even if it exists (EXPERIMENTAL)'),
        ),
    ]
    + commands.diffwsopts
    + commands.walkopts
    + commands.formatteropts,
    'synopsis': _(b'[-r REV] [-f] [-a] [-u] [-d] [-n] [-c] [-l] FILE...'),
    'inferrepo': True,
}


def fastannotate(ui, repo, *pats, **opts):
    """show changeset information by line for each file

    List changes in files, showing the revision id responsible for each line.

    This command is useful for discovering when a change was made and by whom.

    By default this command prints revision numbers. If you include --file,
    --user, or --date, the revision number is suppressed unless you also
    include --number. The default format can also be customized by setting
    fastannotate.defaultformat.

    Returns 0 on success.

    .. container:: verbose

        This command uses an implementation different from the vanilla annotate
        command, which may produce slightly different (while still reasonable)
        outputs for some cases.

        Unlike the vanilla anootate, fastannotate follows rename regardless of
        the existence of --file.

        For the best performance when running on a full repo, use -c, -l,
        avoid -u, -d, -n. Use --linear and --no-content to make it even faster.

        For the best performance when running on a shallow (remotefilelog)
        repo, avoid --linear, --no-follow, or any diff options. As the server
        won't be able to populate annotate cache when non-default options
        affecting results are used.
    """
    if not pats:
        raise error.Abort(_(b'at least one filename or pattern is required'))

    # performance hack: filtered repo can be slow. unfilter by default.
    if ui.configbool(b'fastannotate', b'unfilteredrepo'):
        repo = repo.unfiltered()

    opts = pycompat.byteskwargs(opts)

    rev = opts.get(b'rev', b'.')
    rebuild = opts.get(b'rebuild', False)

    diffopts = patch.difffeatureopts(
        ui, opts, section=b'annotate', whitespace=True
    )
    aopts = facontext.annotateopts(
        diffopts=diffopts,
        followmerge=not opts.get(b'linear', False),
        followrename=not opts.get(b'no_follow', False),
    )

    if not any(
        opts.get(s)
        for s in [b'user', b'date', b'file', b'number', b'changeset']
    ):
        # default 'number' for compatibility. but fastannotate is more
        # efficient with "changeset", "line-number" and "no-content".
        for name in ui.configlist(
            b'fastannotate', b'defaultformat', [b'number']
        ):
            opts[name] = True

    ui.pager(b'fastannotate')
    template = opts.get(b'template')
    if template == b'json':
        formatter = faformatter.jsonformatter(ui, repo, opts)
    else:
        formatter = faformatter.defaultformatter(ui, repo, opts)
    showdeleted = opts.get(b'deleted', False)
    showlines = not bool(opts.get(b'no_content'))
    showpath = opts.get(b'file', False)

    # find the head of the main (master) branch
    master = ui.config(b'fastannotate', b'mainbranch') or rev

    # paths will be used for prefetching and the real annotating
    paths = list(_matchpaths(repo, rev, pats, opts, aopts))

    # for client, prefetch from the server
    if util.safehasattr(repo, 'prefetchfastannotate'):
        repo.prefetchfastannotate(paths)

    for path in paths:
        result = lines = existinglines = None
        while True:
            try:
                with facontext.annotatecontext(repo, path, aopts, rebuild) as a:
                    result = a.annotate(
                        rev,
                        master=master,
                        showpath=showpath,
                        showlines=(showlines and not showdeleted),
                    )
                    if showdeleted:
                        existinglines = {(l[0], l[1]) for l in result}
                        result = a.annotatealllines(
                            rev, showpath=showpath, showlines=showlines
                        )
                break
            except (faerror.CannotReuseError, faerror.CorruptedFileError):
                # happens if master moves backwards, or the file was deleted
                # and readded, or renamed to an existing name, or corrupted.
                if rebuild:  # give up since we have tried rebuild already
                    raise
                else:  # try a second time rebuilding the cache (slow)
                    rebuild = True
                    continue

        if showlines:
            result, lines = result

        formatter.write(result, lines, existinglines=existinglines)
    formatter.end()


_newopts = set()
_knownopts = {
    opt[1].replace(b'-', b'_')
    for opt in (fastannotatecommandargs['options'] + commands.globalopts)
}


def _annotatewrapper(orig, ui, repo, *pats, **opts):
    """used by wrapdefault"""
    # we need this hack until the obsstore has 0.0 seconds perf impact
    if ui.configbool(b'fastannotate', b'unfilteredrepo'):
        repo = repo.unfiltered()

    # treat the file as text (skip the isbinary check)
    if ui.configbool(b'fastannotate', b'forcetext'):
        opts['text'] = True

    # check if we need to do prefetch (client-side)
    rev = opts.get('rev')
    if util.safehasattr(repo, 'prefetchfastannotate') and rev is not None:
        paths = list(_matchpaths(repo, rev, pats, pycompat.byteskwargs(opts)))
        repo.prefetchfastannotate(paths)

    return orig(ui, repo, *pats, **opts)


def registercommand():
    """register the fastannotate command"""
    name = b'fastannotate|fastblame|fa'
    command(name, helpbasic=True, **fastannotatecommandargs)(fastannotate)


def wrapdefault():
    """wrap the default annotate command, to be aware of the protocol"""
    extensions.wrapcommand(commands.table, b'annotate', _annotatewrapper)


@command(
    b'debugbuildannotatecache',
    [(b'r', b'rev', b'', _(b'build up to the specific revision'), _(b'REV'))]
    + commands.walkopts,
    _(b'[-r REV] FILE...'),
)
def debugbuildannotatecache(ui, repo, *pats, **opts):
    """incrementally build fastannotate cache up to REV for specified files

    If REV is not specified, use the config 'fastannotate.mainbranch'.

    If fastannotate.client is True, download the annotate cache from the
    server. Otherwise, build the annotate cache locally.

    The annotate cache will be built using the default diff and follow
    options and lives in '.hg/fastannotate/default'.
    """
    opts = pycompat.byteskwargs(opts)
    rev = opts.get(b'REV') or ui.config(b'fastannotate', b'mainbranch')
    if not rev:
        raise error.Abort(
            _(b'you need to provide a revision'),
            hint=_(b'set fastannotate.mainbranch or use --rev'),
        )
    if ui.configbool(b'fastannotate', b'unfilteredrepo'):
        repo = repo.unfiltered()
    ctx = scmutil.revsingle(repo, rev)
    m = scmutil.match(ctx, pats, opts)
    paths = list(ctx.walk(m))
    if util.safehasattr(repo, 'prefetchfastannotate'):
        # client
        if opts.get(b'REV'):
            raise error.Abort(_(b'--rev cannot be used for client'))
        repo.prefetchfastannotate(paths)
    else:
        # server, or full repo
        progress = ui.makeprogress(_(b'building'), total=len(paths))
        for i, path in enumerate(paths):
            progress.update(i)
            with facontext.annotatecontext(repo, path) as actx:
                try:
                    if actx.isuptodate(rev):
                        continue
                    actx.annotate(rev, rev)
                except (faerror.CannotReuseError, faerror.CorruptedFileError):
                    # the cache is broken (could happen with renaming so the
                    # file history gets invalidated). rebuild and try again.
                    ui.debug(
                        b'fastannotate: %s: rebuilding broken cache\n' % path
                    )
                    actx.rebuild()
                    try:
                        actx.annotate(rev, rev)
                    except Exception as ex:
                        # possibly a bug, but should not stop us from building
                        # cache for other files.
                        ui.warn(
                            _(
                                b'fastannotate: %s: failed to '
                                b'build cache: %r\n'
                            )
                            % (path, ex)
                        )
        progress.complete()



# repair.py - functions for repository repair for mercurial
#
# Copyright 2005, 2006 Chris Mason <mason@suse.com>
# Copyright 2007 Olivia Mackall
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import absolute_import

import errno

from .i18n import _
from .node import (
    hex,
    short,
)
from . import (
    bundle2,
    changegroup,
    discovery,
    error,
    exchange,
    obsolete,
    obsutil,
    pathutil,
    phases,
    pycompat,
    requirements,
    scmutil,
    util,
)
from .utils import (
    hashutil,
    stringutil,
    urlutil,
)


def backupbundle(
    repo, bases, heads, node, suffix, compress=True, obsolescence=True
):
    """create a bundle with the specified revisions as a backup"""

    backupdir = b"strip-backup"
    vfs = repo.vfs
    if not vfs.isdir(backupdir):
        vfs.mkdir(backupdir)

    # Include a hash of all the nodes in the filename for uniqueness
    allcommits = repo.set(b'%ln::%ln', bases, heads)
    allhashes = sorted(c.hex() for c in allcommits)
    totalhash = hashutil.sha1(b''.join(allhashes)).digest()
    name = b"%s/%s-%s-%s.hg" % (
        backupdir,
        short(node),
        hex(totalhash[:4]),
        suffix,
    )

    cgversion = changegroup.localversion(repo)
    comp = None
    if cgversion != b'01':
        bundletype = b"HG20"
        if compress:
            comp = b'BZ'
    elif compress:
        bundletype = b"HG10BZ"
    else:
        bundletype = b"HG10UN"

    outgoing = discovery.outgoing(repo, missingroots=bases, ancestorsof=heads)
    contentopts = {
        b'cg.version': cgversion,
        b'obsolescence': obsolescence,
        b'phases': True,
    }
    return bundle2.writenewbundle(
        repo.ui,
        repo,
        b'strip',
        name,
        bundletype,
        outgoing,
        contentopts,
        vfs,
        compression=comp,
    )


def _collectfiles(repo, striprev):
    """find out the filelogs affected by the strip"""
    files = set()

    for x in pycompat.xrange(striprev, len(repo)):
        files.update(repo[x].files())

    return sorted(files)


def _collectrevlog(revlog, striprev):
    _, brokenset = revlog.getstrippoint(striprev)
    return [revlog.linkrev(r) for r in brokenset]


def _collectbrokencsets(repo, files, striprev):
    """return the changesets which will be broken by the truncation"""
    s = set()

    for revlog in manifestrevlogs(repo):
        s.update(_collectrevlog(revlog, striprev))
    for fname in files:
        s.update(_collectrevlog(repo.file(fname), striprev))

    return s


def strip(ui, repo, nodelist, backup=True, topic=b'backup'):
    # This function requires the caller to lock the repo, but it operates
    # within a transaction of its own, and thus requires there to be no current
    # transaction when it is called.
    if repo.currenttransaction() is not None:
        raise error.ProgrammingError(b'cannot strip from inside a transaction')

    # Simple way to maintain backwards compatibility for this
    # argument.
    if backup in [b'none', b'strip']:
        backup = False

    repo = repo.unfiltered()
    repo.destroying()
    vfs = repo.vfs
    # load bookmark before changelog to avoid side effect from outdated
    # changelog (see repo._refreshchangelog)
    repo._bookmarks
    cl = repo.changelog

    # TODO handle undo of merge sets
    if isinstance(nodelist, bytes):
        nodelist = [nodelist]
    striplist = [cl.rev(node) for node in nodelist]
    striprev = min(striplist)

    files = _collectfiles(repo, striprev)
    saverevs = _collectbrokencsets(repo, files, striprev)

    # Some revisions with rev > striprev may not be descendants of striprev.
    # We have to find these revisions and put them in a bundle, so that
    # we can restore them after the truncations.
    # To create the bundle we use repo.changegroupsubset which requires
    # the list of heads and bases of the set of interesting revisions.
    # (head = revision in the set that has no descendant in the set;
    #  base = revision in the set that has no ancestor in the set)
    tostrip = set(striplist)
    saveheads = set(saverevs)
    for r in cl.revs(start=striprev + 1):
        if any(p in tostrip for p in cl.parentrevs(r)):
            tostrip.add(r)

        if r not in tostrip:
            saverevs.add(r)
            saveheads.difference_update(cl.parentrevs(r))
            saveheads.add(r)
    saveheads = [cl.node(r) for r in saveheads]

    # compute base nodes
    if saverevs:
        descendants = set(cl.descendants(saverevs))
        saverevs.difference_update(descendants)
    savebases = [cl.node(r) for r in saverevs]
    stripbases = [cl.node(r) for r in tostrip]

    stripobsidx = obsmarkers = ()
    if repo.ui.configbool(b'devel', b'strip-obsmarkers'):
        obsmarkers = obsutil.exclusivemarkers(repo, stripbases)
    if obsmarkers:
        stripobsidx = [
            i for i, m in enumerate(repo.obsstore) if m in obsmarkers
        ]

    newbmtarget, updatebm = _bookmarkmovements(repo, tostrip)

    backupfile = None
    node = nodelist[-1]
    if backup:
        backupfile = _createstripbackup(repo, stripbases, node, topic)
    # create a changegroup for all the branches we need to keep
    tmpbundlefile = None
    if saveheads:
        # do not compress temporary bundle if we remove it from disk later
        #
        # We do not include obsolescence, it might re-introduce prune markers
        # we are trying to strip.  This is harmless since the stripped markers
        # are already backed up and we did not touched the markers for the
        # saved changesets.
        tmpbundlefile = backupbundle(
            repo,
            savebases,
            saveheads,
            node,
            b'temp',
            compress=False,
            obsolescence=False,
        )

    with ui.uninterruptible():
        try:
            with repo.transaction(b"strip") as tr:
                # TODO this code violates the interface abstraction of the
                # transaction and makes assumptions that file storage is
                # using append-only files. We'll need some kind of storage
                # API to handle stripping for us.
                oldfiles = set(tr._offsetmap.keys())
                oldfiles.update(tr._newfiles)

                tr.startgroup()
                cl.strip(striprev, tr)
                stripmanifest(repo, striprev, tr, files)

                for fn in files:
                    repo.file(fn).strip(striprev, tr)
                tr.endgroup()

                entries = tr.readjournal()

                for file, troffset in entries:
                    if file in oldfiles:
                        continue
                    with repo.svfs(file, b'a', checkambig=True) as fp:
                        fp.truncate(troffset)
                    if troffset == 0:
                        repo.store.markremoved(file)

                deleteobsmarkers(repo.obsstore, stripobsidx)
                del repo.obsstore
                repo.invalidatevolatilesets()
                repo._phasecache.filterunknown(repo)

            if tmpbundlefile:
                ui.note(_(b"adding branch\n"))
                f = vfs.open(tmpbundlefile, b"rb")
                gen = exchange.readbundle(ui, f, tmpbundlefile, vfs)
                # silence internal shuffling chatter
                maybe_silent = (
                    repo.ui.silent()
                    if not repo.ui.verbose
                    else util.nullcontextmanager()
                )
                with maybe_silent:
                    tmpbundleurl = b'bundle:' + vfs.join(tmpbundlefile)
                    txnname = b'strip'
                    if not isinstance(gen, bundle2.unbundle20):
                        txnname = b"strip\n%s" % urlutil.hidepassword(
                            tmpbundleurl
                        )
                    with repo.transaction(txnname) as tr:
                        bundle2.applybundle(
                            repo, gen, tr, source=b'strip', url=tmpbundleurl
                        )
                f.close()

            with repo.transaction(b'repair') as tr:
                bmchanges = [(m, repo[newbmtarget].node()) for m in updatebm]
                repo._bookmarks.applychanges(repo, tr, bmchanges)

            # remove undo files
            for undovfs, undofile in repo.undofiles():
                try:
                    undovfs.unlink(undofile)
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        ui.warn(
                            _(b'error removing %s: %s\n')
                            % (
                                undovfs.join(undofile),
                                stringutil.forcebytestr(e),
                            )
                        )

        except:  # re-raises
            if backupfile:
                ui.warn(
                    _(b"strip failed, backup bundle stored in '%s'\n")
                    % vfs.join(backupfile)
                )
            if tmpbundlefile:
                ui.warn(
                    _(b"strip failed, unrecovered changes stored in '%s'\n")
                    % vfs.join(tmpbundlefile)
                )
                ui.warn(
                    _(
                        b"(fix the problem, then recover the changesets with "
                        b"\"hg unbundle '%s'\")\n"
                    )
                    % vfs.join(tmpbundlefile)
                )
            raise
        else:
            if tmpbundlefile:
                # Remove temporary bundle only if there were no exceptions
                vfs.unlink(tmpbundlefile)

    repo.destroyed()
    # return the backup file path (or None if 'backup' was False) so
    # extensions can use it
    return backupfile


def softstrip(ui, repo, nodelist, backup=True, topic=b'backup'):
    """perform a "soft" strip using the archived phase"""
    tostrip = [c.node() for c in repo.set(b'sort(%ln::)', nodelist)]
    if not tostrip:
        return None

    backupfile = None
    if backup:
        node = tostrip[0]
        backupfile = _createstripbackup(repo, tostrip, node, topic)

    newbmtarget, updatebm = _bookmarkmovements(repo, tostrip)
    with repo.transaction(b'strip') as tr:
        phases.retractboundary(repo, tr, phases.archived, tostrip)
        bmchanges = [(m, repo[newbmtarget].node()) for m in updatebm]
        repo._bookmarks.applychanges(repo, tr, bmchanges)
    return backupfile


def _bookmarkmovements(repo, tostrip):
    # compute necessary bookmark movement
    bm = repo._bookmarks
    updatebm = []
    for m in bm:
        rev = repo[bm[m]].rev()
        if rev in tostrip:
            updatebm.append(m)
    newbmtarget = None
    # If we need to move bookmarks, compute bookmark
    # targets. Otherwise we can skip doing this logic.
    if updatebm:
        # For a set s, max(parents(s) - s) is the same as max(heads(::s - s)),
        # but is much faster
        newbmtarget = repo.revs(b'max(parents(%ld) - (%ld))', tostrip, tostrip)
        if newbmtarget:
            newbmtarget = repo[newbmtarget.first()].node()
        else:
            newbmtarget = b'.'
    return newbmtarget, updatebm


def _createstripbackup(repo, stripbases, node, topic):
    # backup the changeset we are about to strip
    vfs = repo.vfs
    cl = repo.changelog
    backupfile = backupbundle(repo, stripbases, cl.heads(), node, topic)
    repo.ui.status(_(b"saved backup bundle to %s\n") % vfs.join(backupfile))
    repo.ui.log(
        b"backupbundle", b"saved backup bundle to %s\n", vfs.join(backupfile)
    )
    return backupfile


def safestriproots(ui, repo, nodes):
    """return list of roots of nodes where descendants are covered by nodes"""
    torev = repo.unfiltered().changelog.rev
    revs = {torev(n) for n in nodes}
    # tostrip = wanted - unsafe = wanted - ancestors(orphaned)
    # orphaned = affected - wanted
    # affected = descendants(roots(wanted))
    # wanted = revs
    revset = b'%ld - ( ::( (roots(%ld):: and not _phase(%s)) -%ld) )'
    tostrip = set(repo.revs(revset, revs, revs, phases.internal, revs))
    notstrip = revs - tostrip
    if notstrip:
        nodestr = b', '.join(sorted(short(repo[n].node()) for n in notstrip))
        ui.warn(
            _(b'warning: orphaned descendants detected, not stripping %s\n')
            % nodestr
        )
    return [c.node() for c in repo.set(b'roots(%ld)', tostrip)]


class stripcallback(object):
    """used as a transaction postclose callback"""

    def __init__(self, ui, repo, backup, topic):
        self.ui = ui
        self.repo = repo
        self.backup = backup
        self.topic = topic or b'backup'
        self.nodelist = []

    def addnodes(self, nodes):
        self.nodelist.extend(nodes)

    def __call__(self, tr):
        roots = safestriproots(self.ui, self.repo, self.nodelist)
        if roots:
            strip(self.ui, self.repo, roots, self.backup, self.topic)


def delayedstrip(ui, repo, nodelist, topic=None, backup=True):
    """like strip, but works inside transaction and won't strip irreverent revs

    nodelist must explicitly contain all descendants. Otherwise a warning will
    be printed that some nodes are not stripped.

    Will do a backup if `backup` is True. The last non-None "topic" will be
    used as the backup topic name. The default backup topic name is "backup".
    """
    tr = repo.currenttransaction()
    if not tr:
        nodes = safestriproots(ui, repo, nodelist)
        return strip(ui, repo, nodes, backup=backup, topic=topic)
    # transaction postclose callbacks are called in alphabet order.
    # use '\xff' as prefix so we are likely to be called last.
    callback = tr.getpostclose(b'\xffstrip')
    if callback is None:
        callback = stripcallback(ui, repo, backup=backup, topic=topic)
        tr.addpostclose(b'\xffstrip', callback)
    if topic:
        callback.topic = topic
    callback.addnodes(nodelist)


def stripmanifest(repo, striprev, tr, files):
    for revlog in manifestrevlogs(repo):
        revlog.strip(striprev, tr)


def manifestrevlogs(repo):
    yield repo.manifestlog.getstorage(b'')
    if scmutil.istreemanifest(repo):
        # This logic is safe if treemanifest isn't enabled, but also
        # pointless, so we skip it if treemanifest isn't enabled.
        for t, unencoded, encoded, size in repo.store.datafiles():
            if unencoded.startswith(b'meta/') and unencoded.endswith(
                b'00manifest.i'
            ):
                dir = unencoded[5:-12]
                yield repo.manifestlog.getstorage(dir)


def rebuildfncache(ui, repo):
    """Rebuilds the fncache file from repo history.

    Missing entries will be added. Extra entries will be removed.
    """
    repo = repo.unfiltered()

    if requirements.FNCACHE_REQUIREMENT not in repo.requirements:
        ui.warn(
            _(
                b'(not rebuilding fncache because repository does not '
                b'support fncache)\n'
            )
        )
        return

    with repo.lock():
        fnc = repo.store.fncache
        fnc.ensureloaded(warn=ui.warn)

        oldentries = set(fnc.entries)
        newentries = set()
        seenfiles = set()

        progress = ui.makeprogress(
            _(b'rebuilding'), unit=_(b'changesets'), total=len(repo)
        )
        for rev in repo:
            progress.update(rev)

            ctx = repo[rev]
            for f in ctx.files():
                # This is to minimize I/O.
                if f in seenfiles:
                    continue
                seenfiles.add(f)

                i = b'data/%s.i' % f
                d = b'data/%s.d' % f

                if repo.store._exists(i):
                    newentries.add(i)
                if repo.store._exists(d):
                    newentries.add(d)

        progress.complete()

        if requirements.TREEMANIFEST_REQUIREMENT in repo.requirements:
            # This logic is safe if treemanifest isn't enabled, but also
            # pointless, so we skip it if treemanifest isn't enabled.
            for dir in pathutil.dirs(seenfiles):
                i = b'meta/%s/00manifest.i' % dir
                d = b'meta/%s/00manifest.d' % dir

                if repo.store._exists(i):
                    newentries.add(i)
                if repo.store._exists(d):
                    newentries.add(d)

        addcount = len(newentries - oldentries)
        removecount = len(oldentries - newentries)
        for p in sorted(oldentries - newentries):
            ui.write(_(b'removing %s\n') % p)
        for p in sorted(newentries - oldentries):
            ui.write(_(b'adding %s\n') % p)

        if addcount or removecount:
            ui.write(
                _(b'%d items added, %d removed from fncache\n')
                % (addcount, removecount)
            )
            fnc.entries = newentries
            fnc._dirty = True

            with repo.transaction(b'fncache') as tr:
                fnc.write(tr)
        else:
            ui.write(_(b'fncache already up to date\n'))


def deleteobsmarkers(obsstore, indices):
    """Delete some obsmarkers from obsstore and return how many were deleted

    'indices' is a list of ints which are the indices
    of the markers to be deleted.

    Every invocation of this function completely rewrites the obsstore file,
    skipping the markers we want to be removed. The new temporary file is
    created, remaining markers are written there and on .close() this file
    gets atomically renamed to obsstore, thus guaranteeing consistency."""
    if not indices:
        # we don't want to rewrite the obsstore with the same content
        return

    left = []
    current = obsstore._all
    n = 0
    for i, m in enumerate(current):
        if i in indices:
            n += 1
            continue
        left.append(m)

    newobsstorefile = obsstore.svfs(b'obsstore', b'w', atomictemp=True)
    for bytes in obsolete.encodemarkers(left, True, obsstore._version):
        newobsstorefile.write(bytes)
    newobsstorefile.close()
    return n



# extdiff.py - external diff program support for mercurial
#
# Copyright 2006 Vadim Gelfer <vadim.gelfer@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

'''command to allow external programs to compare revisions

The extdiff Mercurial extension allows you to use external programs
to compare revisions, or revision with working directory. The external
diff programs are called with a configurable set of options and two
non-option arguments: paths to directories containing snapshots of
files to compare.

If there is more than one file being compared and the "child" revision
is the working directory, any modifications made in the external diff
program will be copied back to the working directory from the temporary
directory.

The extdiff extension also allows you to configure new diff commands, so
you do not need to type :hg:`extdiff -p kdiff3` always. ::

  [extdiff]
  # add new command that runs GNU diff(1) in 'context diff' mode
  cdiff = gdiff -Nprc5
  ## or the old way:
  #cmd.cdiff = gdiff
  #opts.cdiff = -Nprc5

  # add new command called meld, runs meld (no need to name twice).  If
  # the meld executable is not available, the meld tool in [merge-tools]
  # will be used, if available
  meld =

  # add new command called vimdiff, runs gvimdiff with DirDiff plugin
  # (see http://www.vim.org/scripts/script.php?script_id=102) Non
  # English user, be sure to put "let g:DirDiffDynamicDiffText = 1" in
  # your .vimrc
  vimdiff = gvim -f "+next" \\
            "+execute 'DirDiff' fnameescape(argv(0)) fnameescape(argv(1))"

Tool arguments can include variables that are expanded at runtime::

  $parent1, $plabel1 - filename, descriptive label of first parent
  $child,   $clabel  - filename, descriptive label of child revision
  $parent2, $plabel2 - filename, descriptive label of second parent
  $root              - repository root
  $parent is an alias for $parent1.

The extdiff extension will look in your [diff-tools] and [merge-tools]
sections for diff tool arguments, when none are specified in [extdiff].

::

  [extdiff]
  kdiff3 =

  [diff-tools]
  kdiff3.diffargs=--L1 '$plabel1' --L2 '$clabel' $parent $child

If a program has a graphical interface, it might be interesting to tell
Mercurial about it. It will prevent the program from being mistakenly
used in a terminal-only environment (such as an SSH terminal session),
and will make :hg:`extdiff --per-file` open multiple file diffs at once
instead of one by one (if you still want to open file diffs one by one,
you can use the --confirm option).

Declaring that a tool has a graphical interface can be done with the
``gui`` flag next to where ``diffargs`` are specified:

::

  [diff-tools]
  kdiff3.diffargs=--L1 '$plabel1' --L2 '$clabel' $parent $child
  kdiff3.gui = true

You can use -I/-X and list of file or directory names like normal
:hg:`diff` command. The extdiff extension makes snapshots of only
needed files, so running the external diff program will actually be
pretty fast (at least faster than having to compare the entire tree).
'''

from __future__ import absolute_import

import os
import re
import shutil
import stat
import subprocess

from mercurial.i18n import _
from mercurial.node import (
    nullrev,
    short,
)
from mercurial import (
    archival,
    cmdutil,
    encoding,
    error,
    filemerge,
    formatter,
    pycompat,
    registrar,
    scmutil,
    util,
)
from mercurial.utils import (
    procutil,
    stringutil,
)

cmdtable = {}
command = registrar.command(cmdtable)

configtable = {}
configitem = registrar.configitem(configtable)

configitem(
    b'extdiff',
    br'opts\..*',
    default=b'',
    generic=True,
)

configitem(
    b'extdiff',
    br'gui\..*',
    generic=True,
)

configitem(
    b'diff-tools',
    br'.*\.diffargs$',
    default=None,
    generic=True,
)

configitem(
    b'diff-tools',
    br'.*\.gui$',
    generic=True,
)

# Note for extension authors: ONLY specify testedwith = 'ships-with-hg-core' for
# extensions which SHIP WITH MERCURIAL. Non-mainline extensions should
# be specifying the version(s) of Mercurial they are tested with, or
# leave the attribute unspecified.
testedwith = b'ships-with-hg-core'


def snapshot(ui, repo, files, node, tmproot, listsubrepos):
    """snapshot files as of some revision
    if not using snapshot, -I/-X does not work and recursive diff
    in tools like kdiff3 and meld displays too many files."""
    dirname = os.path.basename(repo.root)
    if dirname == b"":
        dirname = b"root"
    if node is not None:
        dirname = b'%s.%s' % (dirname, short(node))
    base = os.path.join(tmproot, dirname)
    os.mkdir(base)
    fnsandstat = []

    if node is not None:
        ui.note(
            _(b'making snapshot of %d files from rev %s\n')
            % (len(files), short(node))
        )
    else:
        ui.note(
            _(b'making snapshot of %d files from working directory\n')
            % (len(files))
        )

    if files:
        repo.ui.setconfig(b"ui", b"archivemeta", False)

        archival.archive(
            repo,
            base,
            node,
            b'files',
            match=scmutil.matchfiles(repo, files),
            subrepos=listsubrepos,
        )

        for fn in sorted(files):
            wfn = util.pconvert(fn)
            ui.note(b'  %s\n' % wfn)

            if node is None:
                dest = os.path.join(base, wfn)

                fnsandstat.append((dest, repo.wjoin(fn), os.lstat(dest)))
    return dirname, fnsandstat


def formatcmdline(
    cmdline,
    repo_root,
    do3way,
    parent1,
    plabel1,
    parent2,
    plabel2,
    child,
    clabel,
):
    # Function to quote file/dir names in the argument string.
    # When not operating in 3-way mode, an empty string is
    # returned for parent2
    replace = {
        b'parent': parent1,
        b'parent1': parent1,
        b'parent2': parent2,
        b'plabel1': plabel1,
        b'plabel2': plabel2,
        b'child': child,
        b'clabel': clabel,
        b'root': repo_root,
    }

    def quote(match):
        pre = match.group(2)
        key = match.group(3)
        if not do3way and key == b'parent2':
            return pre
        return pre + procutil.shellquote(replace[key])

    # Match parent2 first, so 'parent1?' will match both parent1 and parent
    regex = (
        br'''(['"]?)([^\s'"$]*)'''
        br'\$(parent2|parent1?|child|plabel1|plabel2|clabel|root)\1'
    )
    if not do3way and not re.search(regex, cmdline):
        cmdline += b' $parent1 $child'
    return re.sub(regex, quote, cmdline)


def _systembackground(cmd, environ=None, cwd=None):
    """like 'procutil.system', but returns the Popen object directly
    so we don't have to wait on it.
    """
    env = procutil.shellenviron(environ)
    proc = subprocess.Popen(
        procutil.tonativestr(cmd),
        shell=True,
        close_fds=procutil.closefds,
        env=procutil.tonativeenv(env),
        cwd=pycompat.rapply(procutil.tonativestr, cwd),
    )
    return proc


def _runperfilediff(
    cmdline,
    repo_root,
    ui,
    guitool,
    do3way,
    confirm,
    commonfiles,
    tmproot,
    dir1a,
    dir1b,
    dir2,
    rev1a,
    rev1b,
    rev2,
):
    # Note that we need to sort the list of files because it was
    # built in an "unstable" way and it's annoying to get files in a
    # random order, especially when "confirm" mode is enabled.
    waitprocs = []
    totalfiles = len(commonfiles)
    for idx, commonfile in enumerate(sorted(commonfiles)):
        path1a = os.path.join(dir1a, commonfile)
        label1a = commonfile + rev1a
        if not os.path.isfile(path1a):
            path1a = pycompat.osdevnull

        path1b = b''
        label1b = b''
        if do3way:
            path1b = os.path.join(dir1b, commonfile)
            label1b = commonfile + rev1b
            if not os.path.isfile(path1b):
                path1b = pycompat.osdevnull

        path2 = os.path.join(dir2, commonfile)
        label2 = commonfile + rev2

        if confirm:
            # Prompt before showing this diff
            difffiles = _(b'diff %s (%d of %d)') % (
                commonfile,
                idx + 1,
                totalfiles,
            )
            responses = _(
                b'[Yns?]'
                b'$$ &Yes, show diff'
                b'$$ &No, skip this diff'
                b'$$ &Skip remaining diffs'
                b'$$ &? (display help)'
            )
            r = ui.promptchoice(b'%s %s' % (difffiles, responses))
            if r == 3:  # ?
                while r == 3:
                    for c, t in ui.extractchoices(responses)[1]:
                        ui.write(b'%s - %s\n' % (c, encoding.lower(t)))
                    r = ui.promptchoice(b'%s %s' % (difffiles, responses))
            if r == 0:  # yes
                pass
            elif r == 1:  # no
                continue
            elif r == 2:  # skip
                break

        curcmdline = formatcmdline(
            cmdline,
            repo_root,
            do3way=do3way,
            parent1=path1a,
            plabel1=label1a,
            parent2=path1b,
            plabel2=label1b,
            child=path2,
            clabel=label2,
        )

        if confirm or not guitool:
            # Run the comparison program and wait for it to exit
            # before we show the next file.
            # This is because either we need to wait for confirmation
            # from the user between each invocation, or because, as far
            # as we know, the tool doesn't have a GUI, in which case
            # we can't run multiple CLI programs at the same time.
            ui.debug(
                b'running %r in %s\n' % (pycompat.bytestr(curcmdline), tmproot)
            )
            ui.system(curcmdline, cwd=tmproot, blockedtag=b'extdiff')
        else:
            # Run the comparison program but don't wait, as we're
            # going to rapid-fire each file diff and then wait on
            # the whole group.
            ui.debug(
                b'running %r in %s (backgrounded)\n'
                % (pycompat.bytestr(curcmdline), tmproot)
            )
            proc = _systembackground(curcmdline, cwd=tmproot)
            waitprocs.append(proc)

    if waitprocs:
        with ui.timeblockedsection(b'extdiff'):
            for proc in waitprocs:
                proc.wait()


def diffpatch(ui, repo, node1, node2, tmproot, matcher, cmdline):
    template = b'hg-%h.patch'
    # write patches to temporary files
    with formatter.nullformatter(ui, b'extdiff', {}) as fm:
        cmdutil.export(
            repo,
            [repo[node1].rev(), repo[node2].rev()],
            fm,
            fntemplate=repo.vfs.reljoin(tmproot, template),
            match=matcher,
        )
    label1 = cmdutil.makefilename(repo[node1], template)
    label2 = cmdutil.makefilename(repo[node2], template)
    file1 = repo.vfs.reljoin(tmproot, label1)
    file2 = repo.vfs.reljoin(tmproot, label2)
    cmdline = formatcmdline(
        cmdline,
        repo.root,
        # no 3way while comparing patches
        do3way=False,
        parent1=file1,
        plabel1=label1,
        # while comparing patches, there is no second parent
        parent2=None,
        plabel2=None,
        child=file2,
        clabel=label2,
    )
    ui.debug(b'running %r in %s\n' % (pycompat.bytestr(cmdline), tmproot))
    ui.system(cmdline, cwd=tmproot, blockedtag=b'extdiff')
    return 1


def diffrevs(
    ui,
    repo,
    ctx1a,
    ctx1b,
    ctx2,
    matcher,
    tmproot,
    cmdline,
    do3way,
    guitool,
    opts,
):

    subrepos = opts.get(b'subrepos')

    # calculate list of files changed between both revs
    st = ctx1a.status(ctx2, matcher, listsubrepos=subrepos)
    mod_a, add_a, rem_a = set(st.modified), set(st.added), set(st.removed)
    if do3way:
        stb = ctx1b.status(ctx2, matcher, listsubrepos=subrepos)
        mod_b, add_b, rem_b = (
            set(stb.modified),
            set(stb.added),
            set(stb.removed),
        )
    else:
        mod_b, add_b, rem_b = set(), set(), set()
    modadd = mod_a | add_a | mod_b | add_b
    common = modadd | rem_a | rem_b
    if not common:
        return 0

    # Always make a copy of ctx1a (and ctx1b, if applicable)
    # dir1a should contain files which are:
    #   * modified or removed from ctx1a to ctx2
    #   * modified or added from ctx1b to ctx2
    #     (except file added from ctx1a to ctx2 as they were not present in
    #     ctx1a)
    dir1a_files = mod_a | rem_a | ((mod_b | add_b) - add_a)
    dir1a = snapshot(ui, repo, dir1a_files, ctx1a.node(), tmproot, subrepos)[0]
    rev1a = b'' if ctx1a.rev() is None else b'@%d' % ctx1a.rev()
    if do3way:
        # file calculation criteria same as dir1a
        dir1b_files = mod_b | rem_b | ((mod_a | add_a) - add_b)
        dir1b = snapshot(
            ui, repo, dir1b_files, ctx1b.node(), tmproot, subrepos
        )[0]
        rev1b = b'@%d' % ctx1b.rev()
    else:
        dir1b = None
        rev1b = b''

    fnsandstat = []

    # If ctx2 is not the wc or there is >1 change, copy it
    dir2root = b''
    rev2 = b''
    if ctx2.node() is not None:
        dir2 = snapshot(ui, repo, modadd, ctx2.node(), tmproot, subrepos)[0]
        rev2 = b'@%d' % ctx2.rev()
    elif len(common) > 1:
        # we only actually need to get the files to copy back to
        # the working dir in this case (because the other cases
        # are: diffing 2 revisions or single file -- in which case
        # the file is already directly passed to the diff tool).
        dir2, fnsandstat = snapshot(ui, repo, modadd, None, tmproot, subrepos)
    else:
        # This lets the diff tool open the changed file directly
        dir2 = b''
        dir2root = repo.root

    label1a = rev1a
    label1b = rev1b
    label2 = rev2

    if not opts.get(b'per_file'):
        # If only one change, diff the files instead of the directories
        # Handle bogus modifies correctly by checking if the files exist
        if len(common) == 1:
            common_file = util.localpath(common.pop())
            dir1a = os.path.join(tmproot, dir1a, common_file)
            label1a = common_file + rev1a
            if not os.path.isfile(dir1a):
                dir1a = pycompat.osdevnull
            if do3way:
                dir1b = os.path.join(tmproot, dir1b, common_file)
                label1b = common_file + rev1b
                if not os.path.isfile(dir1b):
                    dir1b = pycompat.osdevnull
            dir2 = os.path.join(dir2root, dir2, common_file)
            label2 = common_file + rev2

        # Run the external tool on the 2 temp directories or the patches
        cmdline = formatcmdline(
            cmdline,
            repo.root,
            do3way=do3way,
            parent1=dir1a,
            plabel1=label1a,
            parent2=dir1b,
            plabel2=label1b,
            child=dir2,
            clabel=label2,
        )
        ui.debug(b'running %r in %s\n' % (pycompat.bytestr(cmdline), tmproot))
        ui.system(cmdline, cwd=tmproot, blockedtag=b'extdiff')
    else:
        # Run the external tool once for each pair of files
        _runperfilediff(
            cmdline,
            repo.root,
            ui,
            guitool=guitool,
            do3way=do3way,
            confirm=opts.get(b'confirm'),
            commonfiles=common,
            tmproot=tmproot,
            dir1a=os.path.join(tmproot, dir1a),
            dir1b=os.path.join(tmproot, dir1b) if do3way else None,
            dir2=os.path.join(dir2root, dir2),
            rev1a=rev1a,
            rev1b=rev1b,
            rev2=rev2,
        )

    for copy_fn, working_fn, st in fnsandstat:
        cpstat = os.lstat(copy_fn)
        # Some tools copy the file and attributes, so mtime may not detect
        # all changes.  A size check will detect more cases, but not all.
        # The only certain way to detect every case is to diff all files,
        # which could be expensive.
        # copyfile() carries over the permission, so the mode check could
        # be in an 'elif' branch, but for the case where the file has
        # changed without affecting mtime or size.
        if (
            cpstat[stat.ST_MTIME] != st[stat.ST_MTIME]
            or cpstat.st_size != st.st_size
            or (cpstat.st_mode & 0o100) != (st.st_mode & 0o100)
        ):
            ui.debug(
                b'file changed while diffing. '
                b'Overwriting: %s (src: %s)\n' % (working_fn, copy_fn)
            )
            util.copyfile(copy_fn, working_fn)

    return 1


def dodiff(ui, repo, cmdline, pats, opts, guitool=False):
    """Do the actual diff:

    - copy to a temp structure if diffing 2 internal revisions
    - copy to a temp structure if diffing working revision with
      another one and more than 1 file is changed
    - just invoke the diff for a single file in the working dir
    """

    cmdutil.check_at_most_one_arg(opts, b'rev', b'change')
    revs = opts.get(b'rev')
    from_rev = opts.get(b'from')
    to_rev = opts.get(b'to')
    change = opts.get(b'change')
    do3way = b'$parent2' in cmdline

    if change:
        ctx2 = scmutil.revsingle(repo, change, None)
        ctx1a, ctx1b = ctx2.p1(), ctx2.p2()
    elif from_rev or to_rev:
        repo = scmutil.unhidehashlikerevs(
            repo, [from_rev] + [to_rev], b'nowarn'
        )
        ctx1a = scmutil.revsingle(repo, from_rev, None)
        ctx1b = repo[nullrev]
        ctx2 = scmutil.revsingle(repo, to_rev, None)
    else:
        ctx1a, ctx2 = scmutil.revpair(repo, revs)
        if not revs:
            ctx1b = repo[None].p2()
        else:
            ctx1b = repo[nullrev]

    # Disable 3-way merge if there is only one parent
    if do3way:
        if ctx1b.rev() == nullrev:
            do3way = False

    matcher = scmutil.match(ctx2, pats, opts)

    if opts.get(b'patch'):
        if opts.get(b'subrepos'):
            raise error.Abort(_(b'--patch cannot be used with --subrepos'))
        if opts.get(b'per_file'):
            raise error.Abort(_(b'--patch cannot be used with --per-file'))
        if ctx2.node() is None:
            raise error.Abort(_(b'--patch requires two revisions'))

    tmproot = pycompat.mkdtemp(prefix=b'extdiff.')
    try:
        if opts.get(b'patch'):
            return diffpatch(
                ui, repo, ctx1a.node(), ctx2.node(), tmproot, matcher, cmdline
            )

        return diffrevs(
            ui,
            repo,
            ctx1a,
            ctx1b,
            ctx2,
            matcher,
            tmproot,
            cmdline,
            do3way,
            guitool,
            opts,
        )

    finally:
        ui.note(_(b'cleaning up temp directory\n'))
        shutil.rmtree(tmproot)


extdiffopts = (
    [
        (
            b'o',
            b'option',
            [],
            _(b'pass option to comparison program'),
            _(b'OPT'),
        ),
        (b'r', b'rev', [], _(b'revision (DEPRECATED)'), _(b'REV')),
        (b'', b'from', b'', _(b'revision to diff from'), _(b'REV1')),
        (b'', b'to', b'', _(b'revision to diff to'), _(b'REV2')),
        (b'c', b'change', b'', _(b'change made by revision'), _(b'REV')),
        (
            b'',
            b'per-file',
            False,
            _(b'compare each file instead of revision snapshots'),
        ),
        (
            b'',
            b'confirm',
            False,
            _(b'prompt user before each external program invocation'),
        ),
        (b'', b'patch', None, _(b'compare patches for two revisions')),
    ]
    + cmdutil.walkopts
    + cmdutil.subrepoopts
)


@command(
    b'extdiff',
    [
        (b'p', b'program', b'', _(b'comparison program to run'), _(b'CMD')),
    ]
    + extdiffopts,
    _(b'hg extdiff [OPT]... [FILE]...'),
    helpcategory=command.CATEGORY_FILE_CONTENTS,
    inferrepo=True,
)
def extdiff(ui, repo, *pats, **opts):
    """use external program to diff repository (or selected files)

    Show differences between revisions for the specified files, using
    an external program. The default program used is diff, with
    default options "-Npru".

    To select a different program, use the -p/--program option. The
    program will be passed the names of two directories to compare,
    unless the --per-file option is specified (see below). To pass
    additional options to the program, use -o/--option. These will be
    passed before the names of the directories or files to compare.

    The --from, --to, and --change options work the same way they do for
    :hg:`diff`.

    The --per-file option runs the external program repeatedly on each
    file to diff, instead of once on two directories. By default,
    this happens one by one, where the next file diff is open in the
    external program only once the previous external program (for the
    previous file diff) has exited. If the external program has a
    graphical interface, it can open all the file diffs at once instead
    of one by one. See :hg:`help -e extdiff` for information about how
    to tell Mercurial that a given program has a graphical interface.

    The --confirm option will prompt the user before each invocation of
    the external program. It is ignored if --per-file isn't specified.
    """
    opts = pycompat.byteskwargs(opts)
    program = opts.get(b'program')
    option = opts.get(b'option')
    if not program:
        program = b'diff'
        option = option or [b'-Npru']
    cmdline = b' '.join(map(procutil.shellquote, [program] + option))
    return dodiff(ui, repo, cmdline, pats, opts)


class savedcmd(object):
    """use external program to diff repository (or selected files)

    Show differences between revisions for the specified files, using
    the following program::

        %(path)s

    When two revision arguments are given, then changes are shown
    between those revisions. If only one revision is specified then
    that revision is compared to the working directory, and, when no
    revisions are specified, the working directory files are compared
    to its parent.
    """

    def __init__(self, path, cmdline, isgui):
        # We can't pass non-ASCII through docstrings (and path is
        # in an unknown encoding anyway), but avoid double separators on
        # Windows
        docpath = stringutil.escapestr(path).replace(b'\\\\', b'\\')
        self.__doc__ %= {'path': pycompat.sysstr(stringutil.uirepr(docpath))}
        self._cmdline = cmdline
        self._isgui = isgui

    def __call__(self, ui, repo, *pats, **opts):
        opts = pycompat.byteskwargs(opts)
        options = b' '.join(map(procutil.shellquote, opts[b'option']))
        if options:
            options = b' ' + options
        return dodiff(
            ui, repo, self._cmdline + options, pats, opts, guitool=self._isgui
        )


def _gettooldetails(ui, cmd, path):
    """
    returns following things for a
    ```
    [extdiff]
    <cmd> = <path>
    ```
    entry:

    cmd: command/tool name
    path: path to the tool
    cmdline: the command which should be run
    isgui: whether the tool uses GUI or not

    Reads all external tools related configs, whether it be extdiff section,
    diff-tools or merge-tools section, or its specified in an old format or
    the latest format.
    """
    path = util.expandpath(path)
    if cmd.startswith(b'cmd.'):
        cmd = cmd[4:]
        if not path:
            path = procutil.findexe(cmd)
            if path is None:
                path = filemerge.findexternaltool(ui, cmd) or cmd
        diffopts = ui.config(b'extdiff', b'opts.' + cmd)
        cmdline = procutil.shellquote(path)
        if diffopts:
            cmdline += b' ' + diffopts
        isgui = ui.configbool(b'extdiff', b'gui.' + cmd)
    else:
        if path:
            # case "cmd = path opts"
            cmdline = path
            diffopts = len(pycompat.shlexsplit(cmdline)) > 1
        else:
            # case "cmd ="
            path = procutil.findexe(cmd)
            if path is None:
                path = filemerge.findexternaltool(ui, cmd) or cmd
            cmdline = procutil.shellquote(path)
            diffopts = False
        isgui = ui.configbool(b'extdiff', b'gui.' + cmd)
    # look for diff arguments in [diff-tools] then [merge-tools]
    if not diffopts:
        key = cmd + b'.diffargs'
        for section in (b'diff-tools', b'merge-tools'):
            args = ui.config(section, key)
            if args:
                cmdline += b' ' + args
                if isgui is None:
                    isgui = ui.configbool(section, cmd + b'.gui') or False
                break
    return cmd, path, cmdline, isgui


def uisetup(ui):
    for cmd, path in ui.configitems(b'extdiff'):
        if cmd.startswith(b'opts.') or cmd.startswith(b'gui.'):
            continue
        cmd, path, cmdline, isgui = _gettooldetails(ui, cmd, path)
        command(
            cmd,
            extdiffopts[:],
            _(b'hg %s [OPTION]... [FILE]...') % cmd,
            helpcategory=command.CATEGORY_FILE_CONTENTS,
            inferrepo=True,
        )(savedcmd(path, cmdline, isgui))


# tell hggettext to extract docstrings from these functions:
i18nfunctions = [savedcmd]
